"""Module for base classes for specific observable classes."""
from __future__ import annotations

import abc

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted


class BaseObservables(TransformerMixin, BaseEstimator):
    """
    Abstract base class for observable classes.

    This class defines the interface for observable classes. It uses
    the transformer interface from scikit-learn.
    """

    def __init__(self):
        """
        Initialize a BaseObservables instance.

        Initializes the parent classes with the super function.
        """
        super(BaseObservables, self).__init__()

    @abc.abstractmethod
    def fit(self, X, y=None):
        """
        Abstract method for fitting the observables.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray, optional): The target values.

        Raises:
            NotImplementedError: This method must be overwritten by any child class.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def transform(self, X):
        """
        Abstract method for transforming the data.

        Args:
            X (np.ndarray): The input data.

        Raises:
            NotImplementedError: This method must be overwritten by any child class.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_feature_names(self, input_features=None):
        """
        Abstract method for getting the names of the features.

        Args:
            input_features (list of str, optional): The names of the input features.

        Raises:
            NotImplementedError: This method must be overwritten by any child class.
        """
        raise NotImplementedError

    def inverse(self, y):
        """
        Inverse the transformation.

        Args:
            y (np.ndarray): The transformed data.

        Returns:
            np.ndarray: The original data.

        Raises:
            ValueError: If the shape of the input does not match the expected shape.
        """
        check_is_fitted(self, ["n_input_features_", "measurement_matrix_"])
        if y.ndim == 1:
            y = y.reshape(1, -1)
        if y.shape[1] != self.n_output_features_:
            raise ValueError(
                "Wrong number of input features."
                f"Expected y.shape[1] = {self.n_output_features_}; "
                f"instead y.shape[1] = {y.shape[1]}."
            )

        return y @ self.measurement_matrix_.T

    def __add__(self, other):
        if isinstance(self, ConcatObservables):
            return ConcatObservables(self.observables_list_ + [other])
        else:
            return ConcatObservables([self, other])

    @property
    def size(self):
        check_is_fitted(self)
        return self.n_output_features_


# learned from https://github.com/dynamicslab/pysindy/blob/
# d0d96f4466b9c16cdd349fdc515abe9081e5b2cf/pysindy/feature_library/base.py#L235


class ConcatObservables(BaseObservables):
    """
    This class concatenates two or more `BaseObservables` instances into a single
    `ConcatObservables` instance.

    The concatenated observables are handled in such a way that only the first
    observable with the identity mapping is kept, while the identity mapping in
    the rest is removed. The same applies to observables that are polynomials with
    `include_bias=True`, in which case the bias feature is also removed.

    Args:
        observables_list_ (list, optional): A list of `BaseObservables` instances
        to concatenate. Defaults to None.

    Attributes:
        observables_list_ (list, optional): The list of `BaseObservables` instances
            that were concatenated. Defaults to None.
        include_state (bool): True if a linear feature (i.e., the system state) is
            included. This indicator can help to identify if a redundant linear feature
            can be removed.
        n_input_features_ (int): The dimensionality of the input features, e.g.,
            the system state.
        n_output_features_ (int): The dimensionality of the transformed/output
            features, e.g., the observables.
        n_consumed_samples (int): The number of effective samples. This can be less
            than the total number of samples due to time-delay stacking.
        measurement_matrix_ (numpy.ndarray): This matrix transforms a row feature
            vector to return the system state. Its shape is (n_input_features_,
                n_output_features_).

    Methods:
        fit(X, y=None): Calculates and stores important information such as the
            dimensions of the input and output features, the number of effective
            samples, and the measurement matrix.
        transform(X): Applies the transformation defined by the observables to
            input data.
        get_feature_names(input_features=None): Returns the names of the features
            after transformation.
        inverse(y): Applies the inverse transformation to the transformed data to
            recover the original system state.
    """

    def __init__(self, observables_list_=None):
        """Initializes a ConcatObservables instance.

        Args:
            observables_list_ (list, optional): A list of `BaseObservables` instances.
                If provided, the first observable must have an `include_state`
                attribute. The default value is None.

        Raises:
            AssertionError: If the first observable in `observables_list_` does not have
                an `include_state` attribute.
        """
        super(ConcatObservables, self).__init__()
        self.observables_list_ = observables_list_
        assert hasattr(
            self.observables_list_[0], "include_state"
        ), "first observable must have `include_state' attribute"
        self.include_state = self.observables_list_[0].include_state

    def fit(self, X, y=None):
        """Sets up observable by fitting the model to the data.

        This method fits each observable in the list to the data, determines the
        total number of output features, and sets up the measurement matrix.

        Args:
            X (numpy.ndarray): Measurement data to be fit, with shape (n_samples,
                n_input_features_).
            y (numpy.ndarray, optional): Time-shifted measurement data to be fit.
                Default is None.

        Returns:
            ConcatObservables: A fitted instance of the class.

        Raises:
            AssertionError: If any observable in the list does not have an
                `include_state` attribute, or if the shape of the temporary least
                squares solution does not match the shape of the measurement matrix.
        """

        # first, one must call fit of every observable in the observer list
        # so that n_input_features_ and n_output_features_ are defined
        for obs in self.observables_list_:
            obs.fit(X, y)

        self.n_input_features_ = X.shape[1]

        # total number of output features takes care of redundant identity features
        # for polynomial feature, we will remove the 1 as well if include_bias is true

        first_obs = self.observables_list_[0]
        s = 0
        obs_list_contain_state_counter = 1 if first_obs.include_state else 0
        obs_list_contain_bias_counter = (
            1 if getattr(first_obs, "include_bias", False) else 0
        )
        for obs in self.observables_list_[1:]:
            assert hasattr(obs, "include_state"), (
                "observable Must have `include_state' " "attribute"
            )
            if obs_list_contain_state_counter > 1 and obs.include_state:
                s += obs.n_output_features_ - obs.n_input_features_
            else:
                s += obs.n_output_features_
            if obs_list_contain_bias_counter > 1 and getattr(
                obs, "include_bias", False
            ):
                s -= 1
            obs_list_contain_state_counter += 1 if obs.include_state is True else 0
            obs_list_contain_bias_counter += (
                1 if getattr(obs, "include_bias", False) else 0
            )

        self.n_output_features_ = first_obs.n_output_features_ + s

        # take care of consuming samples in time delay observables: \
        # we will look for the largest delay
        max_n_consumed_samples = 0
        for obs in self.observables_list_:
            if hasattr(obs, "n_consumed_samples"):
                max_n_consumed_samples = max(
                    max_n_consumed_samples, obs.n_consumed_samples
                )
        self.n_consumed_samples = max_n_consumed_samples

        # choosing measurement_matrix
        self.measurement_matrix_ = np.zeros(
            [self.n_input_features_, self.n_output_features_]
        )
        # 1. if any observable has `include_state` == True
        if any([obs.include_state for obs in self.observables_list_]) is True:
            jj = 0
            for i in range(len(self.observables_list_)):
                jcol = self.observables_list_[i].measurement_matrix_.shape[1]
                if self.observables_list_[i].include_state is True:
                    break
                jj += jcol
            self.measurement_matrix_[:, jj : jj + jcol] = self.observables_list_[
                i
            ].measurement_matrix_
        else:
            g = self.transform(X)
            tmp = np.linalg.lstsq(g, X)[0].T
            assert tmp.shape == self.measurement_matrix_.shape
            self.measurement_matrix_ = tmp

        # 1. if first observable does not contain include state but others do
        # then we will use the nearest one's measurement matrix

        # otherwise,

        # C comes from the first observable

        # first_obs_measurement_matrix = self.observables_list_[0].measurement_matrix_
        # self.measurement_matrix_[:first_obs_measurement_matrix.shape[0],
        # :first_obs_measurement_matrix.shape[1],] = first_obs_measurement_matrix

        return self

    def transform(self, X):
        """Evaluate observable at `X`.

        This method checks if the model is fitted and then evaluates the observables
        at the provided data, excluding features that are state or bias based on
        certain conditions.

        Args:
            X (numpy.ndarray): Measurement data to be fit, with shape (n_samples,
                n_input_features_).

        Returns:
            y (numpy.ndarray): Evaluation of observables at `X`, with shape (n_samples,
                n_output_features_).

        Raises:
            NotFittedError: If the model is not fitted yet.
        """

        # for obs in self.observables_list_:
        #     check_is_fitted(obs, "n_consumed_samples_")
        check_is_fitted(self, "n_consumed_samples")
        num_samples_updated = X.shape[0] - self.n_consumed_samples
        first_obs = self.observables_list_[0]
        obs_list_contain_state_counter = 1 if first_obs.include_state else 0
        obs_list_contain_bias_counter = (
            1 if getattr(first_obs, "include_bias", False) else 0
        )
        y_list = [first_obs.transform(X)[-num_samples_updated:, :]]

        # only include those features that are not state
        y_rest_list = []
        for obs in self.observables_list_[1:]:
            if obs_list_contain_state_counter > 1 and obs.include_state:
                y_new = obs.transform(X)[-num_samples_updated:, obs.n_input_features_ :]
            else:
                y_new = obs.transform(X)[-num_samples_updated:, :]
            if obs_list_contain_bias_counter > 1 and getattr(
                obs, "include_bias", False
            ):
                y_new = y_new[:, 1:]
            obs_list_contain_state_counter += 1 if obs.include_state else 0
            obs_list_contain_bias_counter += (
                1 if getattr(obs, "include_bias", False) else 0
            )

            y_rest_list.append(y_new)
        y_list += y_rest_list

        # y_list += [
        #     obs.transform(X)[-num_samples_updated:, obs.n_input_features_ :]
        #     for obs in self.observables_list_[1:]
        # ]
        y = np.hstack(y_list)
        return y

    def get_feature_names(self, input_features=None):
        """Return names of observables.

        This method returns a list of feature names, which are created by
        concatenating the feature names from all observables in the list.

        Args:
            input_features (list of str, optional): Default list is "x0", "x1", ...,
            "xn", where n = n_features. Defaults to None.

        Returns:
            list of str: List of feature names of length n_output_features.

        Raises:
            NotFittedError: If the model is not fitted yet.
        """
        check_is_fitted(self, "n_input_features_")

        concat_feature_names = self.observables_list_[0].get_feature_names()
        for obs in self.observables_list_[1:]:
            if getattr(obs, "include_bias", False):
                concat_feature_names += obs.get_feature_names()[
                    obs.n_input_features_ + 1 :
                ]
            else:
                concat_feature_names += obs.get_feature_names()[obs.n_input_features_ :]
        return concat_feature_names

    def inverse(self, y):
        """Invert the transformation to get system state `x`.

        This function approximately (due to some of them use least-square)
        satisfies :code:`self.inverse(self.transform(x)) == x`.

        Args:
            y (numpy.ndarray): Data to which to apply the inverse.
                Shape must be (n_samples, n_output_features).
                Must have the same number of features as the transformed data.

        Returns:
            numpy.ndarray: Output of inverse map applied to y.
                Shape will be (n_samples, n_input_features).
                In this case, x is identical to y.

        Raises:
            NotFittedError: If the model is not fitted yet.
            ValueError: If the number of features in `y` does not match
                `n_output_features_`.

        """

        check_is_fitted(self, ["n_input_features_", "measurement_matrix_"])
        if y.shape[1] != self.n_output_features_:
            raise ValueError(
                "Wrong number of input features."
                f"Expected y.shape[1] = {self.n_output_features_}; "
                f"instead y.shape[1] = {y.shape[1]}."
            )

        # dim_output_first_obs = self.observables_list_[0].n_output_features_
        x = y @ self.measurement_matrix_.T
        return x



"""module for dmd with control"""
from __future__ import annotations

import numpy as np
from sklearn.utils.validation import check_is_fitted

from ._base import BaseRegressor


class DMDc(BaseRegressor):
    """
    Implements Dynamic Mode Decomposition with Control (DMDc) regressor.

    This class provides an implementation for DMDc, a variant of Dynamic Mode
    Decomposition, which is a dimensionality reduction technique used to analyze
    dynamical systems. The goal of DMDc is to compute matrices A and B that satisfy
    the equation x' = Ax + Bu, where x' is the time-shifted state w.r.t. x and u is
    the control input.

    Attributes:
        svd_rank (int): Rank of SVD for the input space, i.e., the space of `X` and
            input `U`. This determines the dimensionality of the projected state and
            control matrices. Defaults to None.
        svd_output_rank (int): Rank of SVD for the output space, i.e., the space of `Y`.
            Defaults to None.
        input_control_matrix (numpy.ndarray): The known input control matrix B. Defaults
            to None.
        n_samples_ (int): Total number of one step evolution samples.
        n_input_features_ (int): Dimension of input features.
        n_control_features_ (int): Dimension of input control signal.
        coef_ (numpy.ndarray): Weight vectors of the regression problem. Corresponds
            to either [A] or [A,B].
        state_matrix_ (numpy.ndarray): Identified state transition matrix A of the
            underlying system.
        control_matrix_ (numpy.ndarray): Identified control matrix B of the underlying
            system.
        reduced_state_matrix_ (numpy.ndarray): Reduced state transition matrix.
        reduced_control_matrix_ (numpy.ndarray): Reduced control matrix.
        eigenvalues_ (numpy.ndarray): DMD lamda.
        unnormalized_modes (numpy.ndarray): DMD V.
        projection_matrix_ (numpy.ndarray): Projection matrix into low-dimensional
            subspace.
        projection_matrix_output_ (numpy.ndarray): Projection matrix into
            low-dimensional subspace.
        eigenvectors_ (numpy.ndarray): DMD eigenvectors.

    Example:
        >>> import numpy as np
        >>> import pykoopman as pk
        >>> A = np.matrix([[1.5, 0],[0, 0.1]])
        >>> B = np.matrix([[1],[0]])
        >>> x0 = np.array([4,7])
        >>> u = np.array([-4, -2, -1, -0.5, 0, 0.5, 1, 3, 5])
        >>> n = len(u)+1
        >>> x = np.zeros([n,len(x0)])
        >>> x[0,:] = x0
        >>> for i in range(n-1):
        >>>     x[i+1,:] = A.dot(x[i,:]) + B.dot(u[np.newaxis,i])
        >>> X1 = x[:-1,:]
        >>> X2 = x[1:,:]
        >>> C = u[:,np.newaxis]
        >>> DMDc = pk.regression.DMDc(svd_rank=3, input_control_matrix=B)
        >>> model = pk.Koopman(regressor=DMDc)
        >>> model.fit(x,C)
        >>> Aest = model.A
        >>> Best = model.B
        >>> print(Aest)
        >>> np.allclose(A,Aest)
        [[ 1.50000000e+00 -1.36609474e-17]
         [-1.58023594e-17  1.00000000e-01]]
        True
    """

    def __init__(self, svd_rank=None, svd_output_rank=None, input_control_matrix=None):
        """
        Initialize a DMDc class object.

        Args:
            svd_rank (int, optional): Rank of SVD for the input space. This determines
                the dimensionality of the projected state and control matrices.
                Defaults to None.
            svd_output_rank (int, optional): Rank of SVD for the output space.
                Defaults to None.
            input_control_matrix (numpy.ndarray, optional): The known input control
                matrix B. Defaults to None.

        Raises:
            ValueError: If svd_rank is not an integer.
            ValueError: If svd_output_rank is not an integer.
            ValueError: If input_control_matrix is not a numpy array.
        """
        self.svd_rank = svd_rank
        self.svd_output_rank = svd_output_rank
        self._input_control_matrix = input_control_matrix

    def fit(self, x, y=None, u=None, dt=None):
        """
        Fit the DMDc model to the provided data.

        Parameters
        ----------
        x : numpy.ndarray, shape (n_samples, n_features)
            Measurement data to be fit.

        y : numpy.ndarray, shape (n_samples, n_features), default=None
            Measurement data output to be fitted.

        u : numpy.ndarray, shape (n_samples, n_control_features), optional, default=None
            Time series of external actuation/control.

        dt : float, optional
            Time interval between `X` and `Y`

        Returns
        -------
        self: returns a fitted ``DMDc`` instance
        """
        self.n_samples_, self.n_input_features_ = x.shape

        if y is None:
            X1 = x[:-1, :]
            X2 = x[1:, :]
        else:
            X1 = x
            X2 = y

        if u.ndim == 1:
            if len(u) > X1.shape[0]:
                u = u[:-1]
            C = u[np.newaxis, :]
        else:
            if u.shape[0] > X1.shape[0]:
                u = u[:-1, :]
            C = u
        self.n_control_features_ = C.shape[1]

        if self.svd_rank is None:
            self.svd_rank = self.n_input_features_ + self.n_control_features_
            if self.svd_output_rank is None:
                self.svd_output_rank = self.n_input_features_
        else:
            if self.svd_output_rank is None:
                self.svd_output_rank = self.svd_rank

        rout = self.svd_output_rank
        r = self.svd_rank

        if self._input_control_matrix is None:
            self._fit_unknown_B(X1, X2, C, r, rout)
        else:
            self._fit_known_B(X1, X2, C, r)

        return self

    def _fit_unknown_B(self, X1, X2, C, r, rout):
        """
        Fits the DMDc model when the control matrix B is unknown. It computes
        the state matrix `A` and control matrix `B` using the Dynamic Mode
        Decomposition with control (DMDc) algorithm.

        Args:
            X1 (numpy.ndarray): The state matrix at time t.
            X2 (numpy.ndarray): The state matrix at time t+1.
            C (numpy.ndarray): The control input matrix.
            r (int): Rank for truncation of singular value decomposition.
            rout (int): Rank for truncation of singular value decomposition on X2
                transpose.

        Returns:
            None. Updates the instance variables _state_matrix_, _control_matrix_,
            _coef_, _eigenvectors_, _eigenvalues_, _ur, _tmp_compute_psi,
            _unnormalized_modes.

        Raises:
            ValueError: If the dimensions of X1, X2, and C are not compatible.
        """

        assert rout <= r
        Omega = np.vstack([X1.T, C.T])

        # SVD of input space
        U, s, Vh = np.linalg.svd(Omega, full_matrices=False)
        Ur = U[:, 0:r]
        Sr = np.diag(s[0:r])
        Vr = Vh[0:r, :].T

        Uhat, _, _ = np.linalg.svd(X2.T, full_matrices=False)
        Uhatr = Uhat[:, 0:rout]

        U1 = Ur[: self.n_input_features_, :]
        U2 = Ur[self.n_input_features_ :, :]

        # this is reduced A_r
        self._state_matrix_ = Uhatr.T @ X2.T @ Vr @ np.linalg.inv(Sr) @ U1.T @ Uhatr
        self._control_matrix_ = Uhatr.T @ X2.T @ Vr @ np.linalg.inv(Sr) @ U2.T

        # self._state_matrix_ = self._reduced_state_matrix_
        # self._control_matrix_ = self._reduced_control_matrix_
        # self._state_matrix_ = Uhatr @ self._reduced_state_matrix_ @ Uhatr.T
        # self._control_matrix_ = Uhatr @ self._reduced_control_matrix_

        # pack [A full, B full] as self.coef_
        self._coef_ = np.concatenate(
            (self._state_matrix_, self._control_matrix_), axis=1
        )

        # self._projection_matrix_ = Ur
        # self._projection_matrix_output_ = Uhatr

        # eigenvectors, lamda
        [self._eigenvalues_, self._eigenvectors_] = np.linalg.eig(self._state_matrix_)

        # Koopman modes V
        self._unnormalized_modes = Uhatr @ self._eigenvectors_
        self._ur = Uhatr
        self._tmp_compute_psi = np.linalg.inv(self._eigenvectors_) @ Uhatr.T

    def _fit_known_B(self, X1, X2, C, r):
        """
        Fits the DMDc model when the control matrix B is known. It computes
        the state matrix `A` using the Dynamic Mode Decomposition with control
        (DMDc) algorithm.

        Args:
            X1 (numpy.ndarray): The state matrix at time t.
            X2 (numpy.ndarray): The state matrix at time t+1.
            C (numpy.ndarray): The control input matrix.
            r (int): Rank for truncation of singular value decomposition.

        Returns:
            None. Updates the instance variables _state_matrix_, _coef_,
            _eigenvectors_, _eigenvalues_, _ur, _tmp_compute_psi, _unnormalized_modes.

        Raises:
            ValueError: If the dimensions of X1, X2, and C are not compatible.
        """
        if self.n_input_features_ in self._input_control_matrix.shape is False:
            raise TypeError("Control vector/matrix B has wrong shape.")
        if self._input_control_matrix.shape[1] == self.n_input_features_:
            self._input_control_matrix = self._input_control_matrix.T
        if self._input_control_matrix.shape[1] != self.n_control_features_:
            raise TypeError(
                "The control matrix B must have the same "
                "number of inputs as the control variable u."
            )

        U, s, Vh = np.linalg.svd(X1.T, full_matrices=False)
        Ur = U[:, :r]
        sr = s[:r]
        Vhr = Vh[:r, :]

        self._state_matrix_ = np.linalg.multi_dot(
            [
                Ur.T,
                X2.T - self._input_control_matrix @ C.T,
                Vhr.T,
                np.diag(np.reciprocal(sr)),
            ]
        )
        self._control_matrix_ = Ur.T @ self._input_control_matrix
        # self._state_matrix_ = Ur @ self._reduced_state_matrix_ @ Ur.T

        self._coef_ = np.concatenate(
            (self._state_matrix_, self.control_matrix_), axis=1
        )
        # self._coef_ = Ur @ self._state_matrix_ @ Ur.T
        # self._projection_matrix_ = Ur
        # self._projection_matrix_output_ = Ur

        # Compute , eigenvectors, lamda
        [self._eigenvalues_, self._eigenvectors_] = np.linalg.eig(self._state_matrix_)

        # Koopman V
        self._unnormalized_modes = Ur @ self._eigenvectors_
        self._ur = Ur
        self._tmp_compute_psi = np.linalg.inv(self._eigenvectors_) @ Ur.T

        # compute psi
        # self.C = np.linalg.inv(self._eigenvectors_) @ Ur.T

    def predict(self, x, u):
        """
        Predicts the future state of the system based on the current state and the
        current value of control input, using the fitted DMDc model.

        Args:
            x (numpy.ndarray): The current state of the system.
            u (numpy.ndarray): The current value of the input.

        Returns:
            numpy.ndarray: The predicted future state of the system.

        Raises:
            NotFittedError: If the model is not fitted, raise this error to prevent
                misuse of the model.
        """
        check_is_fitted(self, "coef_")
        if x.ndim == 1:
            x = x.reshape(1, -1)
        if u.ndim == 1:
            u = u.reshape(1, -1)
        # y = self.coef_ @ np.vstack([x.reshape(1, -1).T, u.reshape(1, -1).T])
        y = (
            x @ self.ur @ self.state_matrix_.T @ self.ur.T
            + u @ self.control_matrix_.T @ self.ur.T
        )
        # y = x @ self.state_matrix_.T + u @ self.control_matrix_.T
        # y = y.T
        return y

    def _compute_phi(self, x_col):
        """
        Returns the transformed matrix `phi(x)` given `x`.

        The method takes a column vector or a 1-D numpy array and computes its
        transformation using the `_ur` matrix. If the input `x_col` is a 1-D array,
        it reshapes it into a column vector before the computation.

        Args:
            x_col (numpy.ndarray): A column vector or a 1-D numpy array
                representing `x`.

        Returns:
            numpy.ndarray: The transformed matrix `phi(x)`.
        """
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        phi = self._ur.T @ x_col
        return phi

    def _compute_psi(self, x_col):
        """
        Returns `psi(x)` given `x`

        Args:
            x: numpy.ndarray, shape (n_samples, n_features)
                Measurement data upon which to compute psi values.

        Returns
            phi : numpy.ndarray, shape (n_samples, n_input_features_) value of
            Koopman psi at x
        """

        # compute psi - one column if x is a row
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        psi = self._tmp_compute_psi @ x_col
        return psi

    @property
    def coef_(self):
        """
        The weight vectors of the regression problem.

        This method checks if the regressor is fitted before returning the coefficient.

        Returns:
            numpy.ndarray: The coefficient matrix.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_coef_")
        return self._coef_

    @property
    def state_matrix_(self):
        """
        The DMD state transition matrix.

        This method checks if the regressor is fitted before returning the state matrix.

        Returns:
            numpy.ndarray: The state transition matrix.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_state_matrix_")
        return self._state_matrix_

    @property
    def control_matrix_(self):
        check_is_fitted(self, "_control_matrix_")
        return self._control_matrix_

    # @property
    # def reduced_state_matrix_(self):
    #     check_is_fitted(self, "_reduced_state_matrix_")
    #     return self._reduced_state_matrix_
    #
    # @property
    # def reduced_control_matrix_(self):
    #     check_is_fitted(self, "_reduced_control_matrix_")
    #     return self._reduced_control_matrix_

    @property
    def eigenvectors_(self):
        """
        The identified Koopman eigenvectors.

        This method checks if the regressor is fitted before returning the eigenvectors.

        Returns:
            numpy.ndarray: The Koopman eigenvectors.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_eigenvectors_")
        return self._eigenvectors_

    @property
    def eigenvalues_(self):
        """
        The identified Koopman eigenvalues.

        This method checks if the regressor is fitted before returning the eigenvalues.

        Returns:
            numpy.ndarray: The Koopman eigenvalues.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_eigenvalues_")
        return self._eigenvalues_

    @property
    def unnormalized_modes(self):
        """
        The raw DMD V with each column as one DMD mode.

        This method checks if the regressor is fitted before returning the unnormalized
            modes.

        Returns:
            numpy.ndarray: The unnormalized modes.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_unnormalized_modes")
        return self._unnormalized_modes

    @property
    def ur(self):
        """
        The left singular vectors 'U'.

        This method checks if the regressor is fitted before returning 'U'.

        Returns:
            numpy.ndarray: The left singular vectors 'U'.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_ur")
        return self._ur

    @property
    def input_control_matrix(self):
        return self._input_control_matrix



"""module for example dynamics data"""
from __future__ import annotations

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from scipy.linalg import orth


def drss(
    n=2, p=2, m=2, p_int_first=0.1, p_int_others=0.01, p_repeat=0.05, p_complex=0.5
):
    """
    Create a discrete-time, random, stable, linear state space model.

    Args:
        n (int, optional): Number of states. Default is 2.
        p (int, optional): Number of control inputs. Default is 2.
        m (int, optional): Number of output measurements.
            If m=0, C becomes the identity matrix, so that y=x. Default is 2.
        p_int_first (float, optional): Probability of an integrator as the first pole.
            Default is 0.1.
        p_int_others (float, optional): Probability of other integrators beyond the
            first. Default is 0.01.
        p_repeat (float, optional): Probability of repeated roots. Default is 0.05.
        p_complex (float, optional): Probability of complex roots. Default is 0.5.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]: A tuple containing the
        state transition matrix (A), control matrix (B), and measurement matrix (C).

        A (numpy.ndarray): State transition matrix of shape (n, n).
        B (numpy.ndarray): Control matrix of shape (n, p).
        C (numpy.ndarray): Measurement matrix of shape (m, n). If m = 0, C is the
            identity matrix.

    """

    # Number of integrators
    nint = int(
        (np.random.rand(1) < p_int_first) + sum(np.random.rand(n - 1) < p_int_others)
    )
    # Number of repeated roots
    nrepeated = int(np.floor(sum(np.random.rand(n - nint) < p_repeat) / 2))
    # Number of complex roots
    ncomplex = int(
        np.floor(sum(np.random.rand(n - nint - 2 * nrepeated, 1) < p_complex) / 2)
    )
    nreal = n - nint - 2 * nrepeated - 2 * ncomplex

    # Random poles
    rep = 2 * np.random.rand(nrepeated) - 1
    if ncomplex != 0:
        mag = np.random.rand(ncomplex)
        cplx = np.zeros(ncomplex, dtype=complex)
        for i in range(ncomplex):
            cplx[i] = mag[i] * np.exp(complex(0, np.pi * np.random.rand(1)))
        re = np.real(cplx)
        im = np.imag(cplx)

    # Generate random state space model
    A = np.zeros((n, n))
    if ncomplex != 0:
        for i in range(0, ncomplex):
            A[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = np.array(
                [[re[i], im[i]], [-im[i], re[i]]]
            )

    if 2 * ncomplex < n:
        list_poles = []
        if nint:
            list_poles = np.append(list_poles, np.ones(nint))
        if rep:
            list_poles = np.append(list_poles, rep)
            list_poles = np.append(list_poles, rep)
        if nreal:
            list_poles = np.append(list_poles, 2 * np.random.rand(nreal) - 1)

        A[2 * ncomplex :, 2 * ncomplex :] = np.diag(list_poles)

    T = orth(np.random.rand(n, n))
    A = np.transpose(T) @ (A @ T)

    # control matrix
    B = np.random.randn(n, p)
    # mask for nonzero entries in B
    mask = np.random.rand(B.shape[0], B.shape[1])
    B = np.squeeze(np.multiply(B, [(mask < 0.75) != 0]))

    # Measurement matrix
    if m == 0:
        C = np.identity(n)
    else:
        C = np.random.randn(m, n)
        mask = np.random.rand(C.shape[0], C.shape[1])
        C = np.squeeze(C * [(mask < 0.75) != 0])

    return A, B, C


def advance_linear_system(x0, u, n, A=None, B=None, C=None):
    """
    Simulate the linear system dynamics for a given number of steps.

    Args:
        x0 (numpy.ndarray): Initial state vector of shape (n,).
        u (numpy.ndarray): Control input array of shape (p,) or (p, n-1).
            If 1-dimensional, it will be converted to a row vector.
        n (int): Number of steps to simulate.
        A (numpy.ndarray, optional): State transition matrix of shape (n, n).
            If not provided, it defaults to None.
        B (numpy.ndarray, optional): Control matrix of shape (n, p).
            If not provided, it defaults to None.
        C (numpy.ndarray, optional): Measurement matrix of shape (m, n).
            If not provided, it defaults to None.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray]: A tuple containing the state trajectory
        (x) and the output trajectory (y).

        x (numpy.ndarray): State trajectory of shape (n, len(x0)).
        y (numpy.ndarray): Output trajectory of shape (n, C.shape[0]).

    """
    if C is None:
        C = np.identity(len(x0))
    if u.ndim == 1:
        u = u[np.newaxis, :]

    y = np.zeros([n, C.shape[0]])
    x = np.zeros([n, len(x0)])
    x[0, :] = x0
    y[0, :] = C.dot(x[0, :])
    for i in range(n - 1):
        x[i + 1, :] = A.dot(x[i, :]) + B.dot(u[:, i])
        y[i + 1, :] = C.dot(x[i + 1, :])
    return x, y


def vdp_osc(t, x, u):
    """
    Compute the dynamics of the Van der Pol oscillator.

    Args:
        t (float): Time.
        x (numpy.ndarray): State vector of shape (2,).
        u (float): Control input.

    Returns:
        numpy.ndarray: Updated state vector of shape (2,).

    """
    y = np.zeros(x.shape)
    y[0, :] = 2 * x[1, :]
    y[1, :] = -0.8 * x[0, :] + 2 * x[1, :] - 10 * (x[0, :] ** 2) * x[1, :] + u
    return y


def rk4(t, x, u, _dt=0.01, func=vdp_osc):
    """
    Perform a 4th order Runge-Kutta integration.

    Args:
        t (float): Time.
        x (numpy.ndarray): State vector of shape (2,).
        u (float): Control input.
        _dt (float, optional): Time step. Defaults to 0.01.
        func (function, optional): Function defining the dynamics. Defaults to vdp_osc.

    Returns:
        numpy.ndarray: Updated state vector of shape (2,).

    """
    # 4th order Runge-Kutta
    k1 = func(t, x, u)
    k2 = func(t, x + k1 * _dt / 2, u)
    k3 = func(t, x + k2 * _dt / 2, u)
    k4 = func(t, x + k1 * _dt, u)
    return x + (_dt / 6) * (k1 + 2 * k2 + 2 * k3 + k4)


def square_wave(step):
    """
    Generate a square wave with a period of 60 time steps.

    Args:
        step (int): Current time step.

    Returns:
        float: Square wave value at the given time step.

    """
    return (-1.0) ** (round(step / 30.0))


def sine_wave(step):
    """
    Generate a sine wave with a period of 60 time steps.

    Args:
        step (int): Current time step.

    Returns:
        float: Sine wave value at the given time step.

    """
    return np.sin(round(step / 30.0))


def lorenz(x, t, sigma=10, beta=8 / 3, rho=28):
    """
    Compute the derivative of the Lorenz system at a given state.

    Args:
        x (list): Current state of the Lorenz system [x, y, z].
        t (float): Current time.
        sigma (float, optional): Parameter sigma. Default is 10.
        beta (float, optional): Parameter beta. Default is 8/3.
        rho (float, optional): Parameter rho. Default is 28.

    Returns:
        list: Derivative of the Lorenz system [dx/dt, dy/dt, dz/dt].

    """
    return [
        sigma * (x[1] - x[0]),
        x[0] * (rho - x[2]) - x[1],
        x[0] * x[1] - beta * x[2],
    ]


def rev_dvdp(t, x, u=0, dt=0.1):
    """
    Reverse dynamics of the Van der Pol oscillator.

    Args:
        t (float): Time.
        x (numpy.ndarray): Current state of the system [x1, x2].
        u (float, optional): Input. Default is 0.
        dt (float, optional): Time step. Default is 0.1.

    Returns:
        numpy.ndarray: Updated state of the system [x1', x2'].

    """
    return np.array(
        [
            x[0, :] - x[1, :] * dt,
            x[1, :] + (x[0, :] - x[1, :] + x[0, :] ** 2 * x[1, :]) * dt,
        ]
    )


class Linear2Ddynamics:
    def __init__(self):
        """
        Initializes a Linear2Ddynamics object.

        """
        self.n_states = 2  # Number of states

    def linear_map(self, x):
        """
        Applies the linear mapping to the input state.

        Args:
            x (numpy.ndarray): Input state.

        Returns:
            numpy.ndarray: Resulting mapped state.

        """
        return np.array([[0.8, -0.05], [0, 0.7]]) @ x

    def collect_data(self, x, n_int, n_traj):
        """
        Collects data by integrating the linear dynamics.

        Args:
            x (numpy.ndarray): Initial state.
            n_int (int): Number of integration steps.
            n_traj (int): Number of trajectories.

        Returns:
            numpy.ndarray: Input data.
            numpy.ndarray: Output data.

        """
        # Init
        X = np.zeros((self.n_states, n_int * n_traj))
        Y = np.zeros((self.n_states, n_int * n_traj))

        # Integrate
        for step in range(n_int):
            y = self.linear_map(x)
            X[:, (step) * n_traj : (step + 1) * n_traj] = x
            Y[:, (step) * n_traj : (step + 1) * n_traj] = y
            x = y

        return X, Y

    def visualize_modes(self, x, phi, eigvals, order=None):
        """
        Visualizes the modes of the linear dynamics.

        Args:
            x (numpy.ndarray): State data.
            phi (numpy.ndarray): Eigenvectors.
            eigvals (numpy.ndarray): Eigenvalues.
            order (list, optional): Order of the modes to visualize. Default is None.

        """
        n_modes = min(10, phi.shape[1])
        fig, axs = plt.subplots(2, n_modes, figsize=(3 * n_modes, 6))
        if order is None:
            index_list = range(n_modes)
        else:
            index_list = order
        j = 0
        for i in index_list:
            axs[0, j].scatter(
                x[0, :],
                x[1, :],
                c=np.real(phi[:, i]),
                marker="o",
                cmap=plt.get_cmap("jet"),
            )
            axs[1, j].scatter(
                x[0, :],
                x[1, :],
                c=np.imag(phi[:, i]),
                marker="o",
                cmap=plt.get_cmap("jet"),
            )
            axs[0, j].set_title(r"$\lambda$=" + "{:2.3f}".format(eigvals[i]))
            j += 1


class torus_dynamics:
    """
    Sparse dynamics in Fourier space on torus.

    Attributes:
        n_states (int): Number of states.
        sparsity (int): Degree of sparsity.
        freq_max (int): Maximum frequency.
        noisemag (float): Magnitude of noise.

    Methods:
        __init__(self, n_states=128, sparsity=5, freq_max=15, noisemag=0.0):
            Initializes a torus_dynamics object.

        setup(self):
            Sets up the dynamics.

        advance(self, n_samples, dt=1):
            Advances the continuous-time dynamics without control.

        advance_discrete_time(self, n_samples, dt, u=None):
            Advances the discrete-time dynamics with or without control.

        set_control_matrix_physical(self, B):
            Sets the control matrix in physical space.

        set_control_matrix_fourier(self, Bhat):
            Sets the control matrix in Fourier space.

        set_point_actuator(self, position=None):
            Sets a single point actuator.

        viz_setup(self):
            Sets up the visualization.

        viz_torus(self, ax, x):
            Visualizes the torus dynamics.

        viz_all_modes(self, modes=None):
            Visualizes all modes.

        modes(self):
            Returns the modes of the dynamics.

        B_effective(self):
            Returns the effective control matrix.

    """

    def __init__(self, n_states=128, sparsity=5, freq_max=15, noisemag=0.0):
        """
        Initializes a torus_dynamics object.

        Args:
            n_states (int, optional): Number of states. Default is 128.
            sparsity (int, optional): Degree of sparsity. Default is 5.
            freq_max (int, optional): Maximum frequency. Default is 15.
            noisemag (float, optional): Magnitude of noise. Default is 0.0.

        """
        self.n_states = n_states
        self.sparsity = sparsity
        self.freq_max = freq_max
        self.noisemag = noisemag
        self.setup()

    def setup(self):
        """
        Sets up the dynamics.

        """
        # Initialization in the Fourier space
        xhat = np.zeros((self.n_states, self.n_states), complex)
        # Index of nonzero frequency components
        self.J = np.zeros((self.sparsity, 2), dtype=int)
        IC = np.zeros(self.sparsity)  # Initial condition, real number
        frequencies = np.zeros(self.sparsity)
        damping = np.zeros(self.sparsity)

        IC = np.random.randn(self.sparsity)
        frequencies = np.sqrt(4 * np.random.rand(self.sparsity))
        damping = -np.random.rand(self.sparsity) * 0.1
        for k in range(self.sparsity):
            loopbreak = 0
            while loopbreak != 1:
                self.J[k, 0] = np.ceil(
                    np.random.rand(1) * self.n_states / (self.freq_max + 1)
                )
                self.J[k, 1] = np.ceil(
                    np.random.rand(1) * self.n_states / (self.freq_max + 1)
                )
                if xhat[self.J[k, 0], self.J[k, 1]] == 0.0:
                    loopbreak = 1

            xhat[self.J[k, 0], self.J[k, 1]] = IC[k]

        mask = np.zeros((self.n_states, self.n_states), int)
        for k in range(self.sparsity):
            mask[self.J[k, 0], self.J[k, 1]] = 1

        self.damping = damping
        self.frequencies = frequencies
        self.IC = IC
        self.xhat = xhat
        self.mask = mask

    def advance(self, n_samples, dt=1):
        """
        Advances the continuous-time dynamics without control.

        Args:
            n_samples (int): Number of samples to advance.
            dt (float, optional): Time step. Default is 1.

        """
        print("Evolving continuous-time dynamics without control.")
        self.n_samples = n_samples
        self.dt = dt

        # Initilization
        # In physical space
        self.X = np.ndarray((self.n_states**2, self.n_samples))
        # In Fourier space
        self.Xhat = np.ndarray((self.n_states**2, self.n_samples), complex)
        self.time_vector = np.zeros(self.n_samples)

        # if self.noisemag != 0:
        #     self.XhatClean = np.ndarray((self.n_states**2, self.n_samples), complex)
        #     self.XClean = np.ndarray((self.n_states**2, self.n_samples))

        for step in range(self.n_samples):
            t = step * self.dt
            self.time_vector[step] = t
            xhat = np.zeros((self.n_states, self.n_states), complex)
            for k in range(self.sparsity):
                xhat[self.J[k, 0], self.J[k, 1]] = (
                    np.exp((self.damping[k] + 1j * 2 * np.pi * self.frequencies[k]) * t)
                    * self.IC[k]
                )

            if self.noisemag != 0:
                self.XhatClean[:, step] = xhat.reshape(self.n_states**2)
                xClean = np.real(np.fft.ifft2(xhat))
                self.XClean[:, step] = xClean.reshape(self.n_states**2)

            # xRMS = np.sqrt(np.mean(xhat.reshape((self.n_states**2,1))**2))
            # xhat = xhat + self.noisemag*xRMS\
            #           *np.random.randn(xhat.shape[0],xhat.shape[1]) \
            #         + 1j*self.noisemag*xRMS \
            #         *np.random.randn(xhat.shape[0],xhat.shape[1])
            self.Xhat[:, step] = xhat.reshape(self.n_states**2)
            x = np.real(np.fft.ifft2(xhat))
            self.X[:, step] = x.reshape(self.n_states**2)

    def advance_discrete_time(self, n_samples, dt, u=None):
        """
        Advances the discrete-time dynamics with or without control.

        Args:
            n_samples (int): Number of samples to advance.
            dt (float): Time step.
            u (array-like, optional): Control input. Default is None.

        """
        print("Evolving discrete-time dynamics with or without control.")
        if u is None:
            self.n_control_features_ = 0
            self.U = np.zeros(n_samples)
            self.U = self.U[np.newaxis, :]
            print("No control input provided. Evolving unforced system.")
        else:
            if u.ndim == 1:
                if len(u) > n_samples:
                    u = u[:-1]
                self.U = u[np.newaxis, :]
            elif u.ndim == 2:
                if u.shape[0] > n_samples:
                    u = u[:-1, :]
                self.U = u
            self.n_control_features_ = self.U.shape[1]

        if not hasattr(self, "B"):
            B = np.zeros((self.n_states, self.n_states))
            print(B.shape)
            self.set_control_matrix_physical(B)
            print("Control matrix is not set. Continue with unforced system.")

        self.n_samples = n_samples
        self.dt = dt

        # Initilization
        # In physical space
        self.X = np.ndarray((self.n_states**2, self.n_samples))
        # In Fourier space
        self.Xhat = np.ndarray((self.n_states**2, self.n_samples), complex)
        self.time_vector = np.zeros(self.n_samples)

        # Set initial condition
        xhat0 = np.zeros((self.n_states, self.n_states), complex)
        for k in range(self.sparsity):
            xhat0[self.J[k, 0], self.J[k, 1]] = self.IC[k]
        self.Xhat[:, 0] = xhat0.reshape(self.n_states**2)
        x0 = np.real(np.fft.ifft2(xhat0))
        self.X[:, 0] = x0.reshape(self.n_states**2)

        for step in range(1, self.n_samples, 1):
            t = step * self.dt
            self.time_vector[step] = t
            # self.Xhat[:, step] = np.reshape(self.Bhat * self.U[0,step - 1],\
            #                   self.n_states ** 2)
            # xhat = self.Xhat[:,step].reshape(self.n_states,self.n_states)
            # xhat_prev = \
            # self.Xhat[:, step - 1].reshape(self.n_states, self.n_states)

            # forced torus dynamics linearly evolve in the spectral space, sparsely
            xhat = np.array((self.n_states, self.n_states), complex)
            xhat = self.Xhat[:, step].reshape(self.n_states, self.n_states)
            xhat_prev = self.Xhat[:, step - 1].reshape(self.n_states, self.n_states)
            for k in range(self.sparsity):
                xhat[self.J[k, 0], self.J[k, 1]] = (
                    np.exp(
                        (self.damping[k] + 1j * 2 * np.pi * self.frequencies[k])
                        * self.dt
                    )
                    * xhat_prev[self.J[k, 0], self.J[k, 1]]
                    + self.Bhat[self.J[k, 0], self.J[k, 1]] * self.U[0, step - 1]
                )

            # xhat_prev = self.Xhat[:,step-1].reshape(self.n_states, self.n_states)
            # for k in range(self.sparsity):
            #     xhat[self.J[k,0], self.J[k,1]] += np.exp((self.damping[k] \
            #     + 1j * 2 * np.pi * self.frequencies[k]) * self.dt) \
            #     * xhat_prev[self.J[k,0], self.J[k,1]]

            self.Xhat[:, step] = xhat.reshape(self.n_states**2)
            x = np.real(np.fft.ifft2(xhat))
            self.X[:, step] = x.reshape(self.n_states**2)

    def set_control_matrix_physical(self, B):
        """
        Sets the control matrix in physical space.

        Args:
            B (array-like): Control matrix in physical space.

        """
        if np.allclose(B.shape, np.array([self.n_states, self.n_states])) is False:
            raise TypeError("Control matrix B has wrong shape.")
        self.B = B
        self.Bhat = np.fft.fft2(B)

    def set_control_matrix_fourier(self, Bhat):
        """
        Sets the control matrix in Fourier space.

        Args:
            Bhat (array-like): Control matrix in Fourier space.

        """
        if np.allclose(Bhat.shape, np.array([self.n_states, self.n_states])) is False:
            raise TypeError("Control matrix Bhat has wrong shape.")
        self.Bhat = Bhat
        self.B = np.real(np.fft.ifft2(self.Bhat))

    def set_point_actuator(self, position=None):
        """
        Sets a single point actuator.

        Args:
            position (array-like, optional): Position of the actuator. Default is None.

        """
        if position is None:
            position = np.random.randint(0, self.n_states, 2)
        try:
            for i in range(len(position)):
                position[i] = int(position[i])
        except ValueError:
            print("position was not a valid integer.")

        is_position_in_valid_domain = (position >= 0) & (position < self.n_states)
        if all(is_position_in_valid_domain) is False:
            raise ValueError(
                "Actuator position was not a valid integer inside of domain."
            )

        # Control matrix in physical space (single point actuator)
        B = np.zeros((self.n_states, self.n_states))
        B[position[0], position[1]] = 1
        self.set_control_matrix_physical(B)

    def viz_setup(self):
        """
        Sets up the visualization.

        """
        self.cmap_torus = plt.cm.jet  # bwr #plt.cm.RdYlBu
        self.n_colors = self.n_states
        r1 = 2
        r2 = 1
        [T1, T2] = np.meshgrid(
            np.linspace(0, 2 * np.pi, self.n_states),
            np.linspace(0, 2 * np.pi, self.n_states),
        )
        R = r1 + r2 * np.cos(T2)
        self.Zgrid = r2 * np.sin(T2)
        self.Xgrid = R * np.cos(T1)
        self.Ygrid = R * np.sin(T1)

    def viz_torus(self, ax, x):
        """
        Visualizes the torus dynamics.

        Args:
            ax: Axes object for plotting.
            x (array-like): Dynamics to be visualized.

        Returns:
            surface: Surface plot of the torus dynamics.

        """
        if not hasattr(self, "viz"):
            self.viz_setup()

        norm = mpl.colors.Normalize(vmin=-abs(x).max(), vmax=abs(x).max())
        surface = ax.plot_surface(
            self.Xgrid,
            self.Ygrid,
            self.Zgrid,
            facecolors=self.cmap_torus(norm(x)),
            shade=False,
            rstride=1,
            cstride=1,
        )
        #     m = cm.ScalarMappable(cmap=cmap_torus, norm=norm)
        #     m.set_array([])
        #     plt.colorbar(m)
        # ax.figure.colorbar(surf, ax=ax)
        ax.set_zlim(-3.01, 3.01)
        return surface

    def viz_all_modes(self, modes=None):
        """
        Visualizes all modes.

        Args:
            modes (array-like, optional): Modes to be visualized. Default is None.

        Returns:
            fig: Figure object containing the visualizations.

        """
        if modes is None:
            modes = self.modes

        if not hasattr(self, "viz"):
            self.viz_setup()

        fig = plt.figure(figsize=(20, 10))
        for k in range(self.sparsity):
            ax = plt.subplot2grid((1, self.sparsity), (0, k), projection="3d")
            self.viz_torus(ax, modes[:, k].reshape(self.n_states, self.n_states))
            plt.axis("off")
        return fig

    @property
    def modes(self):
        """
        Returns the modes of the dynamics.

        Returns:
            modes (array-like): Modes of the dynamics.

        """
        modes = np.zeros((self.n_states**2, self.sparsity))

        for k in range(self.sparsity):
            mode_in_fourier = np.zeros((self.n_states, self.n_states))
            mode_in_fourier[self.J[k, 0], self.J[k, 1]] = 1
            modes[:, k] = np.real(
                np.fft.ifft2(mode_in_fourier).reshape(self.n_states**2)
            )

        return modes

    @property
    def B_effective(self):
        """
        Returns the effective control matrix.

        Returns:
            B_effective (array-like): Effective control matrix.

        """
        Bhat_effective = np.zeros((self.n_states, self.n_states), complex)
        for k in range(self.sparsity):
            control_mode = np.zeros((self.n_states, self.n_states), complex)
            control_mode[self.J[k, 0], self.J[k, 1]] = self.Bhat[
                self.J[k, 0], self.J[k, 1]
            ]
            Bhat_effective += control_mode
        B_effective = np.fft.ifft2(Bhat_effective)

        return B_effective


class slow_manifold:
    """
    Represents the slow manifold class.

    Args:
        mu (float, optional): Parameter mu. Default is -0.05.
        la (float, optional): Parameter la. Default is -1.0.
        dt (float, optional): Time step size. Default is 0.01.

    Attributes:
        mu (float): Parameter mu.
        la (float): Parameter la.
        b (float): Value computed from mu and la.
        dt (float): Time step size.
        n_states (int): Number of states.

    Methods:
        sys(t, x, u): Computes the system dynamics.
        output(x): Computes the output based on the state.
        simulate(x0, n_int): Simulates the system dynamics.
        collect_data_continuous(x0): Collects data from continuous-time dynamics.
        collect_data_discrete(x0, n_int): Collects data from discrete-time dynamics.
        visualize_trajectories(t, X, n_traj): Visualizes the trajectories.
        visualize_state_space(X, Y, n_traj): Visualizes the state space.
    """

    def __init__(self, mu=-0.05, la=-1.0, dt=0.01):
        self.mu = mu
        self.la = la
        self.b = self.la / (self.la - 2 * self.mu)
        self.dt = dt
        self.n_states = 2

    def sys(self, t, x, u):
        """
        Computes the system dynamics.

        Args:
            t (float): Time.
            x (array-like): State.
            u (array-like): Control input.

        Returns:
            array-like: Computed system dynamics.

        """
        return np.array([self.mu * x[0, :], self.la * (x[1, :] - x[0, :] ** 2)])

    def output(self, x):
        """
        Computes the output based on the state.

        Args:
            x (array-like): State.

        Returns:
            array-like: Computed output.

        """
        return x[0, :] ** 2 + x[1, :]

    def simulate(self, x0, n_int):
        """
        Simulates the system dynamics.

        Args:
            x0 (array-like): Initial state.
            n_int (int): Number of integration steps.

        Returns:
            array-like: Simulated trajectory.

        """
        n_traj = x0.shape[1]
        x = x0
        u = np.zeros((n_int, 1))
        X = np.zeros((self.n_states, n_int * n_traj))
        for step in range(n_int):
            y = rk4(0, x, u[step, :], self.dt, self.sys)
            X[:, (step) * n_traj : (step + 1) * n_traj] = y
            x = y
        return X

    def collect_data_continuous(self, x0):
        """
        Collects data from continuous-time dynamics.

        Args:
            x0 (array-like): Initial state.

        Returns:
            tuple: Collected data (X, Y).

        """
        n_traj = x0.shape[1]
        u = np.zeros((1, n_traj))
        X = x0
        Y = self.sys(0, x0, u)
        return X, Y

    def collect_data_discrete(self, x0, n_int):
        """
        Collects data from discrete-time dynamics.

        Args:
            x0 (array-like): Initial state.
            n_int (int): Number of integration steps.

        Returns:
            tuple: Collected data (X, Y).

        """
        n_traj = x0.shape[1]
        x = x0
        u = np.zeros((n_int, n_traj))
        X = np.zeros((self.n_states, n_int * n_traj))
        Y = np.zeros((self.n_states, n_int * n_traj))
        for step in range(n_int):
            y = rk4(0, x, u[step, :], self.dt, self.sys)
            X[:, (step) * n_traj : (step + 1) * n_traj] = x
            Y[:, (step) * n_traj : (step + 1) * n_traj] = y
            x = y
        return X, Y

    def visualize_trajectories(self, t, X, n_traj):
        """
        Visualizes the trajectories.

        Args:
            t (array-like): Time vector.
            X (array-like): State trajectories.
            n_traj (int): Number of trajectories.

        """
        fig, axs = plt.subplots(1, 1, tight_layout=True, figsize=(12, 4))
        for traj_idx in range(n_traj):
            x = X[:, traj_idx::n_traj]
            axs.plot(t[0:100], x[1, 0:100], "k")
        axs.set(ylabel=r"$x_2$", xlabel=r"$t$")

    def visualize_state_space(self, X, Y, n_traj):
        """
        Visualizes the state space.

        Args:
            X (array-like): State trajectories.
            Y (array-like): Output trajectories.
            n_traj (int): Number of trajectories.

        """
        fig, axs = plt.subplots(1, 1, tight_layout=True, figsize=(4, 4))
        for traj_idx in range(n_traj):
            axs.plot(
                [X[0, traj_idx::n_traj], Y[0, traj_idx::n_traj]],
                [X[1, traj_idx::n_traj], Y[1, traj_idx::n_traj]],
                "-k",
            )
        axs.set(ylabel=r"$x_2$", xlabel=r"$x_1$")


class forced_duffing:
    """
    Forced Duffing Oscillator.

    dx1/dt = x2
    dx2/dt = -d*x2-alpha*x1-beta*x1^3 + u

    [1] S. Peitz, S. E. Otto, and C. W. Rowley,
    “Data-driven model predictive control using interpolated koopman generators,”
    SIAM J. Appl. Dyn. Syst., vol. 19, no. 3, pp. 2162–2193, Mar. 2020.
    """

    def __init__(self, dt, d, alpha, beta):
        """
        Initializes the Forced Duffing Oscillator.

        Args:
            dt (float): Time step.
            d (float): Damping coefficient.
            alpha (float): Coefficient of x1.
            beta (float): Coefficient of x1^3.
        """
        self.dt = dt
        self.d = d
        self.alpha = alpha
        self.beta = beta
        self.n_states = 2

    def sys(self, t, x, u):
        """
        Defines the system dynamics of the Forced Duffing Oscillator.

        Args:
            t (float): Time.
            x (array-like): State vector.
            u (array-like): Control input.

        Returns:
            array-like: Rate of change of the state vector.
        """
        y = np.array(
            [
                x[1, :],
                -self.d * x[1, :] - self.alpha * x[0, :] - self.beta * x[0, :] ** 3 + u,
            ]
        )
        return y

    def simulate(self, x0, n_int, u):
        """
        Simulates the Forced Duffing Oscillator.

        Args:
            x0 (array-like): Initial state vector.
            n_int (int): Number of time steps.
            u (array-like): Control inputs.

        Returns:
            array-like: State trajectories.
        """
        n_traj = x0.shape[1]
        x = x0
        X = np.zeros((self.n_states, n_int * n_traj))
        for step in range(n_int):
            y = rk4(0, x, u[step, :], self.dt, self.sys)
            X[:, (step) * n_traj : (step + 1) * n_traj] = y
            x = y
        return X

    def collect_data_continuous(self, x0, u):
        """
        Collects continuous data for the Forced Duffing Oscillator.

        Args:
            x0 (array-like): Initial state vector.
            u (array-like): Control inputs.

        Returns:
            tuple: State and output trajectories.
        """
        X = x0
        Y = self.sys(0, x0, u)
        return X, Y

    def collect_data_discrete(self, x0, n_int, u):
        """
        Collects discrete-time data for the Forced Duffing Oscillator.

        Args:
            x0 (array-like): Initial state vector.
            n_int (int): Number of time steps.
            u (array-like): Control inputs.

        Returns:
            tuple: State and output trajectories.
        """
        n_traj = x0.shape[1]
        x = x0
        X = np.zeros((self.n_states, n_int * n_traj))
        Y = np.zeros((self.n_states, n_int * n_traj))
        for step in range(n_int):
            y = rk4(0, x, u[step, :], self.dt, self.sys)
            X[:, (step) * n_traj : (step + 1) * n_traj] = x
            Y[:, (step) * n_traj : (step + 1) * n_traj] = y
            x = y
        return X, Y

    def visualize_trajectories(self, t, X, n_traj):
        """
        Visualizes the state trajectories of the Forced Duffing Oscillator.

        Args:
            t (array-like): Time vector.
            X (array-like): State trajectories.
            n_traj (int): Number of trajectories to visualize.
        """
        fig, axs = plt.subplots(1, 2, tight_layout=True, figsize=(12, 4))
        for traj_idx in range(n_traj):
            x = X[:, traj_idx::n_traj]
            axs[0].plot(t, x[0, :], "k")
            axs[1].plot(t, x[1, :], "b")
        axs[0].set(ylabel=r"$x_1$", xlabel=r"$t$")
        axs[1].set(ylabel=r"$x_2$", xlabel=r"$t$")

    def visualize_state_space(self, X, Y, n_traj):
        """
        Visualizes the state space trajectories of the Forced Duffing Oscillator.

        Args:
            X (array-like): State trajectories.
            Y (array-like): Output trajectories.
            n_traj (int): Number of trajectories to visualize.
        """
        fig, axs = plt.subplots(1, 1, tight_layout=True, figsize=(4, 4))
        for traj_idx in range(n_traj):
            axs.plot(
                [X[0, traj_idx::n_traj], Y[0, traj_idx::n_traj]],
                [X[1, traj_idx::n_traj], Y[1, traj_idx::n_traj]],
                "-k",
            )
        axs.set(ylabel=r"$x_2$", xlabel=r"$x_1$")



from __future__ import annotations

from ._base_analyzer import BaseAnalyzer
from ._ms_pd21 import ModesSelectionPAD21
from ._pruned_koopman import PrunedKoopman

__all__ = ["BaseAnalyzer", "ModesSelectionPAD21", "PrunedKoopman"]



"""module for discrete time Koopman class"""
from __future__ import annotations

from warnings import catch_warnings
from warnings import filterwarnings
from warnings import warn

import numpy as np
from numpy import empty
from pydmd import DMD
from pydmd import DMDBase
from sklearn.base import BaseEstimator
from sklearn.metrics import r2_score
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from .common import validate_input
from .observables import Identity
from .observables import TimeDelay
from .regression import BaseRegressor
from .regression import DMDc
from .regression import EDMDc
from .regression import EnsembleBaseRegressor
from .regression import NNDMD
from .regression import PyDMDRegressor


class Koopman(BaseEstimator):
    """Discrete-Time Koopman class.

    The input-output data is all row-wise if stated elsewhere.
    All of the matrix, are based on column-wise linear system.
    This class is inherited from `pykoopman.regression.BaseEstimator`.

    Args:
        observables: observables object, optional
            (default: `pykoopman.observables.Identity`)
            Map(s) to apply to raw measurement data before estimating the
            Koopman operator.
            Must extend `pykoopman.observables.BaseObservables`.
            The default option, `pykoopman.observables.Identity`, leaves
            the input untouched.

        regressor: regressor object, optional (default: `DMD`)
            The regressor used to learn the Koopman operator from the observables.
            `regressor` can either extend the `pykoopman.regression.BaseRegressor`,
            or `pydmd.DMDBase`.
            In the latter case, the pydmd object must have both a `fit`
            and a `predict` method.

        quiet: boolean, optional (default: False)
            Whether or not warnings should be silenced during fitting.

    Attributes:
        model: sklearn.pipeline.Pipeline
            Internal representation of the forward model.
            Applies the observables and the regressor.

        n_input_features_: int
            Number of input features before computing observables.

        n_output_features_: int
            Number of output features after computing observables.

        n_control_features_: int
            Number of control features used as input to the system.

        time: dictionary
            Time vector properties.
    """

    def __init__(self, observables=None, regressor=None, quiet=False):
        """Constructor for the Koopman class.

        Args:
            observables: observables object, optional
                (default: `pykoopman.observables.Identity`)
                Map(s) to apply to raw measurement data before estimating the
                Koopman operator.
                Must extend `pykoopman.observables.BaseObservables`.
                The default option, `pykoopman.observables.Identity`, leaves
                the input untouched.

            regressor: regressor object, optional (default: `DMD`)
                The regressor used to learn the Koopman operator from the observables.
                `regressor` can either extend the `pykoopman.regression.BaseRegressor`,
                or `pydmd.DMDBase`.
                In the latter case, the pydmd object must have both a `fit`
                and a `predict` method.

            quiet: boolean, optional (default: False)
                Whether or not warnings should be silenced during fitting.
        """
        if observables is None:
            observables = Identity()
        if regressor is None:
            regressor = PyDMDRegressor(DMD(svd_rank=2))  # set default svd rank 2
        if isinstance(regressor, DMDBase):
            regressor = PyDMDRegressor(regressor)
        elif not isinstance(regressor, (BaseRegressor)):
            raise TypeError("Regressor must be from valid class")
        self.observables = observables
        self.regressor = regressor
        self.quiet = quiet

    def fit(self, x, y=None, u=None, dt=1):
        """
        Fit the Koopman model by learning an approximate Koopman operator.

        Args:
            x: numpy.ndarray, shape (n_samples, n_features)
                Measurement data to be fit. Each row should correspond to an example
                and each column a feature. If only x is provided, it is assumed that
                examples are equi-spaced in time (i.e., a uniform timestep is assumed).

            y: numpy.ndarray, shape (n_samples, n_features), optional (default: None)
                Target measurement data to be fit, i.e., it is assumed y = fun(x). Each
                row should correspond to an example and each column a feature. The
                samples in x and y are generally not required to be consecutive and
                equi-spaced.

            u: numpy.ndarray, shape (n_samples, n_control_features), optional (default:
                None) Control/actuation/external parameter data. Each row should
                correspond to one sample and each column a control variable or feature.
                The control variable may be the amplitude of an actuator or an external,
                time-varying parameter. It is assumed that samples in u occur at the
                time instances of the corresponding samples in x,
                e.g., x(t+1) = fun(x(t), u(t)).

            dt: float, optional (default: 1)
                Time step between samples

        Returns:
            self: returns a fit `Koopman` instance
        """
        x = validate_input(x)

        if u is None:
            self.n_control_features_ = 0
        elif not isinstance(self.regressor, DMDc) and not isinstance(
            self.regressor, EDMDc
        ):
            raise ValueError(
                "Control input u was passed, " "but self.regressor is not DMDc or EDMDc"
            )

        if y is None:  # or isinstance(self.regressor, PyDMDRegressor):
            # if there is only 1 trajectory OR regressor is PyDMD
            regressor = self.regressor
        elif isinstance(self.regressor, NNDMD):
            regressor = self.regressor
        else:
            # multiple 1-step-trajectories
            regressor = EnsembleBaseRegressor(
                regressor=self.regressor,
                func=self.observables.transform,
                inverse_func=self.observables.inverse,
            )
            # if x is a list, we need to further change trajectories into 1-step-traj
            if isinstance(x, list):
                x_tmp = []
                y_tmp = []
                for traj_dat in x:
                    x_tmp.append(traj_dat[:-1])
                    y_tmp.append(traj_dat[1:])
                x = np.hstack(x_tmp)
                y = np.hstack(y_tmp)

        steps = [
            ("observables", self.observables),
            ("regressor", regressor),
        ]
        self._pipeline = Pipeline(steps)  # create `model` object using Pipeline

        action = "ignore" if self.quiet else "default"
        with catch_warnings():
            filterwarnings(action, category=UserWarning)
            if u is None:
                self._pipeline.fit(x, y, regressor__dt=dt)
            else:
                self._pipeline.fit(x, y, regressor__u=u, regressor__dt=dt)
            # update the second step with just the regressor, not the
            # EnsembleBaseRegressor
            if isinstance(self._pipeline.steps[1][1], EnsembleBaseRegressor):
                self._pipeline.steps[1] = (
                    self._pipeline.steps[1][0],
                    self._pipeline.steps[1][1].regressor_,
                )

        # pykoopman's n_input/output_features are simply
        # observables's input output features
        # observable's input features are just the number
        # of states. but the output features can be really high
        self.n_input_features_ = self._pipeline.steps[0][1].n_input_features_
        self.n_output_features_ = self._pipeline.steps[0][1].n_output_features_
        if hasattr(self._pipeline.steps[1][1], "n_control_features_"):
            self.n_control_features_ = self._pipeline.steps[1][1].n_control_features_

        # compute amplitudes
        if isinstance(x, list):
            self._amplitudes = None
        elif y is None:
            if hasattr(self.observables, "n_consumed_samples"):
                # g0 = self.observables.transform(
                #     x[0 : 1 + self.observables.n_consumed_samples]
                # )
                self._amplitudes = np.abs(
                    self.psi(x[0 : 1 + self.observables.n_consumed_samples].T)
                )
            else:
                # g0 = self.observables.transform(x[0:1])

                self._amplitudes = np.abs(self.psi(x[0:1].T))
        else:
            self._amplitudes = None

        self.time = {
            "tstart": 0,
            "tend": dt * (self._pipeline.steps[1][1].n_samples_ - 1),
            "dt": dt,
        }

        return self

    def predict(self, x, u=None):
        """
        Predict the state one timestep in the future.

        Args:
            x: numpy.ndarray, shape (n_samples, n_input_features)
                Current state.

            u: numpy.ndarray, shape (n_samples, n_control_features),
                optional (default None)
                Time series of external actuation/control.

        Returns:
            x_next: numpy.ndarray, shape (n_samples, n_input_features)
                Predicted state one timestep in the future.
        """

        check_is_fitted(self, "n_output_features_")
        x_next = self.observables.inverse(self._step(x, u))
        return x_next

    def simulate(self, x0, u=None, n_steps=1):
        """Simulate an initial state forward in time with the learned Koopman model.

        Args:
            x0: numpy.ndarray, shape (n_input_features,) or
                (n_consumed_samples + 1, n_input_features)
                Initial state from which to simulate.
                If using TimeDelay observables, `x0` should contain
                enough examples to compute all required time delays,
                i.e., `n_consumed_samples + 1`.

            u: numpy.ndarray, shape (n_samples, n_control_features),
                optional (default None)
                Time series of external actuation/control.

            n_steps: int, optional (default 1)
                Number of forward steps to be simulated.

        Returns:
            x: numpy.ndarray, shape (n_steps, n_input_features)
                Simulated states.
                Note that `x[0, :]` is one timestep ahead of `x0`.
        """
        check_is_fitted(self, "n_output_features_")
        # Could have an option to only return the end state and not all
        # intermediate states to save memory.

        if x0.ndim == 1:  # handle non-time delay input but 1D accidently
            x0 = x0.reshape(-1, 1)
        elif x0.ndim == 2 and x0.shape[0] > 1:  # handle time delay input
            x0 = x0.T
        else:
            raise TypeError("Check your initial condition shape!")

        # x = empty((n_steps, self.n_input_features_), dtype=self.A.dtype)
        y = empty((n_steps, self.A.shape[0]), dtype=self.W.dtype)

        if u is None:
            # lifted eigen space and move 1 step forward
            y[0] = self.lamda @ self.psi(x0).flatten()

            # iterate in the lifted space
            for k in range(n_steps - 1):
                # tmp = self.W @ self.lamda**(k+1) @ y[0].reshape(-1,1)
                y[k + 1] = self.lamda @ y[k]
            x = np.transpose(self.W @ y.T)
            x = x.astype(self.A.dtype)
        else:
            # lifted space (not eigen)
            y[0] = self.A @ self.phi(x0).flatten() + self.B @ u[0]

            # iterate in the lifted space
            for k in range(n_steps - 1):
                tmp = self.A @ y[k].reshape(-1, 1) + self.B @ u[k + 1].reshape(-1, 1)
                y[k + 1] = tmp.flatten()
            x = np.transpose(self.C @ y.T)
            x = x.astype(self.A.dtype)

        return x

    def get_feature_names(self, input_features=None):
        """Get the names of the individual features constituting the observables.

        Args:
            input_features: list of string, length n_input_features,
                optional (default None)
                String names for input features, if available. By default,
                the names "x0", "x1", ..., "xn_input_features" are used.

        Returns:
            output_feature_names: list of string, length n_output_features
                Output feature names.
        """
        check_is_fitted(self, "n_input_features_")
        return self.observables.get_feature_names(input_features=input_features)

    def _step(self, x, u=None):
        """Map x one timestep forward in the space of observables.

        Args:
            x: numpy.ndarray, shape (n_samples, n_input_features)
                State vectors to be stepped forward.

            u: numpy.ndarray, shape (n_samples, n_control_features),
                optional (default None)
                Time series of external actuation/control.

        Returns:
            X': numpy.ndarray, shape (n_samples, self.n_output_features_)
                Observables one timestep after x.
        """
        check_is_fitted(self, "n_output_features_")

        if u is None or self.n_control_features_ == 0:
            if self.n_control_features_ > 0:
                raise TypeError(
                    "Model was fit using control variables, so u is required"
                )
            elif u is not None:
                warn(
                    "Control variables u were ignored because control variables were"
                    " not used when the model was fit"
                )
            return self._pipeline.predict(X=x)
        else:
            if not isinstance(self.regressor, DMDc) and not isinstance(
                self.regressor, EDMDc
            ):
                raise ValueError(
                    "Control input u was passed, but self.regressor is not DMDc "
                    "or EDMDc"
                )
            return self._pipeline.predict(X=x, u=u)

    def phi(self, x_col):
        """Compute the feature matrix phi(x) given `x_col`.

        Args:
            x_col: numpy.ndarray, shape (n_features, n_samples)
                State vectors to be evaluated for phi.

        Returns:
            phi: numpy.ndarray, shape (n_samples, self.n_output_features_)
                Value of phi evaluated at input `x_col`.
        """
        x = x_col.T
        y = self.observables.transform(x)
        phi = self._pipeline.steps[-1][1]._compute_phi(y.T)
        return phi

    def psi(self, x_col):
        """Compute the Koopman psi(x) given `x_col`.

        Args:
            x_col: numpy.ndarray, shape (n_features, n_samples)
                State vectors to be evaluated for psi.

        Returns:
            eigen_phi: numpy.ndarray, shape (n_samples, self.n_output_features_)
                Value of psi evaluated at input `x_col`.
        """
        x = x_col.T
        y = self.observables.transform(x)
        ephi = self._pipeline.steps[-1][1]._compute_psi(y.T)
        return ephi

    @property
    def A(self):
        """Returns the state transition matrix `A`.

        The state transition matrix A satisfies y' = Ay or y' = Ay + Bu,
        respectively, where y = g(x) and y is a low-rank representation.
        """
        check_is_fitted(self, "_pipeline")
        if isinstance(self.regressor, DMDBase):
            raise ValueError("self.regressor " "has no A!")
        if hasattr(self._pipeline.steps[-1][1], "state_matrix_"):
            return self._pipeline.steps[-1][1].state_matrix_
        else:
            raise ValueError("self.regressor" "has no state_matrix")

    @property
    def B(self):
        """Returns the control matrix `B`.

        The control matrix (or vector) B satisfies y' = Ay + Bu.
        y is the reduced system state.
        """
        check_is_fitted(self, "_pipeline")
        if isinstance(self.regressor, DMDBase):
            raise ValueError("this type of self.regressor has no B")
        return self._pipeline.steps[-1][1].control_matrix_

    @property
    def C(self):
        """Returns the measurement matrix (or vector) C.

        The measurement matrix C satisfies x = C * phi_r.
        """
        check_is_fitted(self, "_pipeline")
        # if not isinstance(self.observables, RadialBasisFunction):
        #     raise ValueError("this type of self.observable has no C")
        # return self._pipeline.steps[0][1].measurement_matrix_

        measure_mat = self._pipeline.steps[0][1].measurement_matrix_
        ur = self._pipeline.steps[-1][1].ur
        C = measure_mat @ ur
        return C

    @property
    def W(self):
        """Returns the Koopman modes."""

        check_is_fitted(self, "_pipeline")
        # return self.C @ self._pipeline.steps[-1][1].unnormalized_modes
        return self.C @ self._pipeline.steps[-1][1].eigenvectors_

    @property
    def _regressor_eigenvectors(self):
        """Returns the eigenvectors of the regressor."""
        check_is_fitted(self, "_pipeline")
        return self._pipeline.steps[-1][1].eigenvectors_

    @property
    def lamda(self):
        """Returns the discrete-time Koopman lambda obtained from spectral
        decomposition."""
        check_is_fitted(self, "_pipeline")
        return np.diag(self._pipeline.steps[-1][1].eigenvalues_)

    @property
    def lamda_array(self):
        """Returns the discrete-time Koopman lambda as an array."""
        check_is_fitted(self, "_pipeline")
        return np.diag(self.lamda) + 0j

    @property
    def continuous_lamda_array(self):
        """Returns the continuous-time Koopman lambda as an array."""
        check_is_fitted(self, "_pipeline")
        return np.log(self.lamda_array) / self.time["dt"]

    @property
    def ur(self):
        """Returns the projection matrix Ur."""
        check_is_fitted(self, "_pipeline")
        return self._pipeline.steps[-1][1].ur

    def validity_check(self, t, x):
        """Perform a validity check of eigenfunctions.

        The validity check tests the linearity of eigenfunctions phi(x(t)) == phi(x(0))
        * exp(lambda*t).

        Args:
            t: numpy.ndarray, shape (n_samples,)
                Time vector.
            x: numpy.ndarray, shape (n_samples, n_input_features)
                State vectors to be checked.

        Returns:
            efun_index: list
                Sorted indices of eigenfunctions based on linearity error.
            linearity_error: list
                Linearity error for each eigenfunction.
        """

        psi = self.psi(x.T)
        omega = np.log(np.diag(self.lamda) + 0j) / self.time["dt"]

        # omega = self.eigenvalues_continuous
        linearity_error = []
        for i in range(self.lamda.shape[0]):
            linearity_error.append(
                np.linalg.norm(psi[i, :] - np.exp(omega[i] * t) * psi[i, 0:1])
            )
        sort_idx = np.argsort(linearity_error)
        efun_index = np.arange(len(linearity_error))[sort_idx]
        linearity_error = [linearity_error[i] for i in sort_idx]
        return efun_index, linearity_error

    def score(self, x, y=None, cast_as_real=True, metric=r2_score, **metric_kws):
        """Score the model predictions for the next timestep.

        Parameters:
            x: numpy.ndarray, shape (n_samples, n_input_features)
                State measurements.
            y: numpy.ndarray, shape (n_samples, n_input_features), optional
                (default None). State measurements one timestep in the future.
            cast_as_real: bool, optional (default True)
                Whether to take the real part of predictions when computing the score.
            metric: callable, optional (default r2_score)
                The metric function used to score the model predictions.
            metric_kws: dict, optional
                Optional parameters to pass to the metric function.

        Returns:
            score: float
                Metric function value for the model predictions at the next timestep.
        """
        check_is_fitted(self, "n_output_features_")
        x = validate_input(x)

        if isinstance(self.observables, TimeDelay):
            n_consumed_samples = self.observables.n_consumed_samples

            # User may pass in too-large
            if y is not None and len(y) == len(x):
                warn(
                    f"The first {n_consumed_samples} entries of y were ignored because "
                    "TimeDelay obesrvables were used."
                )
                y = y[n_consumed_samples:]
        else:
            n_consumed_samples = 0

        if y is None:
            if cast_as_real:
                return metric(
                    x[n_consumed_samples + 1 :].real,
                    self.predict(x[:-1]).real,
                    **metric_kws,
                )
            else:
                return metric(
                    x[n_consumed_samples + 1 :], self.predict(x[:-1]), **metric_kws
                )
        else:
            if cast_as_real:
                return metric(y.real, self.predict(x).real, **metric_kws)
            else:
                return metric(y, self.predict(x), **metric_kws)

    def _observable(self):
        """Returns the observable transformation."""
        return self._pipeline.steps[0][1]

    def _regressor(self):
        """Returns the fitted regressor."""
        # this can access the fitted regressor
        # todo: future we need to figure out a way to do time delay multiple
        #  trajectories DMD
        # my idea is to manually call xN observables then concate the data to let
        # the _regressor.fit to update the model coefficients.
        # call this function with _regressor()
        return self._pipeline.steps[1][1]



"""module for Radial basis function observables"""
from __future__ import annotations

import numpy as np
from numpy import empty
from numpy import random
from sklearn.utils.validation import check_is_fitted

from ..common import validate_input
from ._base import BaseObservables


class RadialBasisFunction(BaseObservables):
    """
    This class represents Radial Basis Functions (RBF) used as observables.
    Observables are formed as RBFs of the state variables, interpreted as new state
    variables.

    For instance, a single state variable :math:`[x(t)]` could be evaluated using
    multiple centers, yielding a new set of observables. This implementation supports
    various types of RBFs including 'gauss', 'thinplate', 'invquad', 'invmultquad',
    and 'polyharmonic'.

    Attributes:
        rbf_type (str): The type of radial basis functions to be used.
        n_centers (int): The number of centers to compute RBF with.
        centers (numpy array): The centers to compute RBF with.
        kernel_width (float): The kernel width for Gaussian RBFs.
        polyharmonic_coeff (float): The polyharmonic coefficient for polyharmonic RBFs.
        include_state (bool): Whether to include the input coordinates as additional
            coordinates in the observable.
        n_input_features_ (int): Number of input features.
        n_output_features_ (int): Number of output features = Number of centers plus
            number of input features.

    Note:
        The implementation is based on the following references:
        - Williams, Matthew O and Kevrekidis, Ioannis G and Rowley, Clarence W
          "A data-driven approximation of the {K}oopman operator: extending dynamic
          mode decomposition."
          Journal of Nonlinear Science 6 (2015): 1307-1346
        - Williams, Matthew O and Rowley, Clarence W and Kevrekidis, Ioannis G
          "A Kernel Approach to Data-Driven {K}oopman Spectral Analysis."
          Journal of Computational Dynamics 2.2 (2015): 247-265
        - Korda, Milan and Mezic, Igor
          "Linear predictors for nonlinear dynamical systems: Koopman operator meets
          model predictive control."
          Automatica 93 (2018): 149-160
    """

    def __init__(
        self,
        rbf_type="gauss",
        n_centers=10,
        centers=None,
        kernel_width=1.0,
        polyharmonic_coeff=1.0,
        include_state=True,
    ):
        super().__init__()
        if type(rbf_type) != str:
            raise TypeError("rbf_type must be a string")
        if type(n_centers) != int:
            raise TypeError("n_centers must be an int")
        if n_centers < 0:
            raise ValueError("n_centers must be a nonnegative int")
        if kernel_width < 0:
            raise ValueError("kernel_width must be a nonnegative float")
        if polyharmonic_coeff < 0:
            raise ValueError("polyharmonic_coeff must be a nonnegative float")
        if rbf_type not in [
            "thinplate",
            "gauss",
            "invquad",
            "invmultquad",
            "polyharmonic",
        ]:
            raise ValueError("rbf_type not of available type")
        if type(include_state) != bool:
            raise TypeError("include_states must be a boolean")
        if centers is not None:
            if int(n_centers) not in centers.shape:
                raise ValueError(
                    "n_centers is not equal to centers.shape[1]. "
                    "centers must be of shape (n_input_features, "
                    "n_centers). "
                )
        self.rbf_type = rbf_type
        self.n_centers = int(n_centers)
        self.centers = centers
        self.kernel_width = kernel_width
        self.polyharmonic_coeff = polyharmonic_coeff
        self.include_state = include_state

    def fit(self, x, y=None):
        """
        Initializes the RadialBasisFunction with specified parameters.

        Args:
            rbf_type (str, optional): The type of radial basis functions to be used.
                Options are: 'gauss', 'thinplate', 'invquad', 'invmultquad',
                'polyharmonic'. Defaults to 'gauss'.
            n_centers (int, optional): The number of centers to compute RBF with.
                Must be a non-negative integer. Defaults to 10.
            centers (numpy array, optional): The centers to compute RBF with.
                If provided, it should have a shape of (n_input_features, n_centers).
                Defaults to None, in which case the centers are uniformly distributed
                over input data.
            kernel_width (float, optional): The kernel width for Gaussian RBFs.
                Must be a non-negative float. Defaults to 1.0.
            polyharmonic_coeff (float, optional): The polyharmonic coefficient for
                polyharmonic RBFs. Must be a non-negative float. Defaults to 1.0.
            include_state (bool, optional): Whether to include the input coordinates
                as additional coordinates in the observable. Defaults to True.

        Raises:
            TypeError: If rbf_type is not a string, n_centers is not an int, or
                include_state is not a bool.
            ValueError: If n_centers, kernel_width or polyharmonic_coeff is negative,
                rbf_type is not of available type, or centers is provided but
                n_centers is not equal to centers.shape[1].
        """
        x = validate_input(x)
        n_samples, n_features = x.shape
        self.n_consumed_samples = 0

        self.n_samples_ = n_samples
        self.n_input_features_ = n_features
        if self.include_state is True:
            self.n_output_features_ = n_features * 1 + self.n_centers
        elif self.include_state is False:
            self.n_output_features_ = self.n_centers

        x = validate_input(x)

        if x.shape[1] != self.n_input_features_:
            raise ValueError(
                "Wrong number of input features. "
                f"Expected x.shape[1] = {self.n_input_features_}; "
                f"instead x.shape[1] = {x.shape[1]}."
            )

        if self.centers is None:
            # Uniformly distributed centers
            self.centers = random.rand(self.n_input_features_, self.n_centers)
            # Change range to range of input features' range
            for feat in range(self.n_input_features_):
                xminmax = self._minmax(x[:, feat])

                # Map to range [0,1]
                self.centers[feat, :] = (
                    self.centers[feat, :] - min(self.centers[feat, :])
                ) / (max(self.centers[feat, :]) - min(self.centers[feat, :]))
                # Scale to input features' range
                self.centers[feat, :] = (
                    self.centers[feat, :] * (xminmax[1] - xminmax[0]) + xminmax[0]
                )

        xlift = self._rbf_lifting(x)
        # self.measurement_matrix_ = x.T @ np.linalg.pinv(xlift.T)
        self.measurement_matrix_ = np.linalg.lstsq(xlift, x)[0].T

        return self

    def transform(self, x):
        """
        Apply radial basis function transformation to the data.

        Args:
            x (array-like): Measurement data to be transformed, with shape (n_samples,
                n_input_features). It is assumed that rows correspond to examples,
                which are not required to be equi-spaced in time or in sequential order.

        Returns:
            array-like: Transformed data, with shape (n_samples, n_output_features).

        Raises:
            NotFittedError: If the 'fit' method has not been called before the
                'transform' method.
            ValueError: If the number of features in 'x' does not match the number of
                input features expected by the transformer.
        """
        check_is_fitted(self, ["n_input_features_", "centers"])
        x = validate_input(x)

        if x.shape[1] != self.n_input_features_:
            raise ValueError(
                "Wrong number of input features. "
                f"Expected x.shape[1] = {self.n_input_features_}; "
                f"instead x.shape[1] = {x.shape[1]}."
            )

        y = self._rbf_lifting(x)
        return y

    def get_feature_names(self, input_features=None):
        """
        Get the names of the output features.

        Args:
            input_features (list of str, optional): String names for input features,
                if available. By default, the names "x0", "x1", ... ,
                "xn_input_features" are used.

        Returns:
            list of str: Output feature names.

        Raises:
            NotFittedError: If the 'fit' method has not been called before the
                'get_feature_names' method.
            ValueError: If the length of 'input_features' does not match the number of
                input features expected by the transformer.
        """

        check_is_fitted(self, "n_input_features_")
        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )

        output_features = []
        if self.include_state is True:
            output_features.extend([f"{xi}(t)" for xi in input_features])
        output_features.extend([f"phi(x(t)-c{i})" for i in range(self.n_centers)])
        return output_features

    def _rbf_lifting(self, x):
        """
        Internal method that performs Radial Basis Function (RBF) transformation.

        Args:
            x (numpy.ndarray): Input data of shape (n_samples, n_input_features)

        Returns:
            y (numpy.ndarray): Transformed data of shape (n_samples, n_output_features)

        Raises:
            ValueError: If 'rbf_type' is not one of the available types.

        Notes:
            This method should not be called directly. It is used internally by the
            'transform' method.
        """
        n_samples = x.shape[0]
        y = empty(
            (n_samples, self.n_output_features_),
            dtype=x.dtype,
        )

        y_index = 0
        if self.include_state is True:
            y[:, : self.n_input_features_] = x
            y_index = self.n_input_features_

        for index_of_center in range(self.n_centers):
            C = self.centers[:, index_of_center]
            r_squared = np.sum((x - C[np.newaxis, :]) ** 2, axis=1)

            match self.rbf_type:
                case "thinplate":
                    y_ = r_squared * np.log(np.sqrt(r_squared))
                    y_[np.isnan(y_)] = 0
                case "gauss":
                    y_ = np.exp(-self.kernel_width**2 * r_squared)
                case "invquad":
                    y_ = np.reciprocal(1 + self.kernel_width**2 * r_squared)
                case "invmultquad":
                    y_ = np.reciprocal(np.sqrt(1 + self.kernel_width**2 * r_squared))
                case "polyharmonic":
                    y_ = r_squared ** (self.polyharmonic_coeff / 2) * np.log(
                        np.sqrt(r_squared)
                    )
                case _:
                    # if none of the above cases match:
                    raise ValueError("provided rbf_type not available")

            y[:, y_index + index_of_center] = y_

        return y

    def _minmax(self, x):
        min_val = min(x)
        max_val = max(x)
        return (min_val, max_val)



"""module for continuous time Koopman class"""
from __future__ import annotations

import numpy as np
from sklearn.utils.validation import check_is_fitted

from .differentiation import Derivative
from .koopman import Koopman


class KoopmanContinuous(Koopman):
    """
    Continuous-time Koopman class.

    Args:
        observables: Observables object, optional
            (default: pykoopman.observables.Identity)
            Map(s) to apply to raw measurement data before
            estimating the Koopman operator. Must extend
            pykoopman.observables.BaseObservables. The default
            option, pykoopman.observables.Identity, leaves the
            input untouched.
        differentiator: Callable, optional
            (default: centered difference)
            Function used to compute numerical derivatives.
            The function must have the call signature
            differentiator(x, t), where x is a 2D numpy ndarray
            of shape (n_samples, n_features) and t is a 1D numpy
            ndarray of shape (n_samples,).
        regressor: Regressor object, optional
            (default: DMD)
            The regressor used to learn the Koopman operator from
            the observables. regressor can either extend
            pykoopman.regression.BaseRegressor, or the
            pydmd.DMDBase class. In the latter case, the pydmd
            object must have both a fit and a predict method.
    """

    def __init__(
        self,
        observables=None,
        differentiator=Derivative(kind="finite_difference", k=1),
        regressor=None,
    ):
        """
        Continuous-time Koopman class.

        Args:
            observables: Observables object, optional
                (default: pykoopman.observables.Identity)
                Map(s) to apply to raw measurement data before
                estimating the Koopman operator. Must extend
                pykoopman.observables.BaseObservables. The default
                option, pykoopman.observables.Identity, leaves the
                input untouched.
            differentiator: Callable, optional
                (default: centered difference)
                Function used to compute numerical derivatives.
                The function must have the call signature
                differentiator(x, t), where x is a 2D numpy ndarray
                of shape (n_samples, n_features) and t is a 1D numpy
                ndarray of shape (n_samples,).
            regressor: Regressor object, optional
                (default: DMD)
                The regressor used to learn the Koopman operator from
                the observables. regressor can either extend
                pykoopman.regression.BaseRegressor, or the
                pydmd.DMDBase class. In the latter case, the pydmd
                object must have both a fit and a predict method.
        """
        super().__init__(observables, regressor)
        self.differentiator = differentiator

    def predict(self, x, dt=0, u=None):
        """
        Predict using continuous-time Koopman model.

        Args:
            x: numpy.ndarray
                State measurements. Each row should correspond to
                the system state at some point in time.
            dt: float, optional (default: 0)
                Time step between measurements. If specified, the
                prediction is made for the given time step in the
                future.
            u: numpy.ndarray, optional (default: None)
                Control input/actuation data. Each row should
                correspond to one sample and each column a control
                variable or feature.

        Returns:
            output: numpy.ndarray
                Predicted state using the continuous-time Koopman
                model. Each row corresponds to the predicted state
                for the corresponding row in x.
        """
        check_is_fitted(self, "_pipeline")

        if u is None:
            ypred = self._pipeline.predict(X=x, t=dt)
        else:
            ypred = self._pipeline.predict(X=x, u=u, t=dt)

        output = self.observables.inverse(ypred)

        return output

    def simulate(self, x, t=0, u=None):
        """
        Simulate continuous-time Koopman model.

        Args:
            x: numpy.ndarray
                Initial state from which to simulate. Each row
                corresponds to the system state at some point in time.
            t: float, optional (default: 0)
                Time at which to simulate the system. If specified,
                the simulation is performed for the given time.
            u: numpy.ndarray, optional (default: None)
                Control input/actuation data. Each row should
                correspond to one sample and each column a control
                variable or feature.

        Returns:
            output: numpy.ndarray
                Simulated states of the system. Each row corresponds
                to the simulated state at a specific time point.
        """
        check_is_fitted(self, "_pipeline")

        if u is None:
            ypred = self._pipeline.predict(X=x, t=t)
        else:
            ypred = self._pipeline.predict(X=x, u=u, t=t)

        output = []
        for k in range(ypred.shape[0]):
            output.append(np.squeeze(self.observables.inverse(ypred[k][np.newaxis, :])))

        return np.array(output)

    def _step(self, x, u=None):
        """
        Placeholder method for step function.

        This method is not implemented in the ContinuousKoopman class
        as there is no explicit step function for continuous-time
        Koopman models.

        Raises:
            NotImplementedError: This method is not implemented
                in the ContinuousKoopman class.
        """
        raise NotImplementedError("ContinuousKoopman does not have a step function.")



"""moduel for Polynomial observables"""
from __future__ import annotations

from itertools import chain
from itertools import combinations
from itertools import combinations_with_replacement as combinations_w_r

import numpy as np
from scipy import sparse
from sklearn.preprocessing import PolynomialFeatures
from sklearn.preprocessing._csr_polynomial_expansion import _csr_polynomial_expansion
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import FLOAT_DTYPES

from ..common import check_array
from ..common import validate_input
from ._base import BaseObservables


class Polynomial(PolynomialFeatures, BaseObservables):
    """
    Polynomial observables.

    This is essentially the `sklearn.preprocessing.PolynomialFeatures` with support for
    complex numbers.

    Args:
        degree (int, optional): The degree of the polynomial features. Default is 2.
        interaction_only (bool, optional): If True, only interaction features are
            produced: features that are products of at most ``degree`` *distinct*
            input features (so not ``x[1] ** 2``, ``x[0] * x[2] ** 3``,
            etc.). Default is False.
        include_bias (bool, optional): If True, then include a bias column, the feature
            in which all polynomial powers are zero (i.e., a column of ones - acts as an
            intercept term in a linear model). Default is True.
        order (str in {'C', 'F'}, optional): Order of output array in the dense case.
            'F' order is faster to compute, but may slow down subsequent estimators.
            Default is 'C'.

    Raises:
        ValueError: If degree is less than 1.
    """

    def __init__(self, degree=2, interaction_only=False, include_bias=True, order="C"):
        """
        Initialize the Polynomial object.

        Args:
            degree (int, optional): The degree of the polynomial features. Default is 2.
            interaction_only (bool, optional): If True, only interaction features are
                produced: features that are products of at most ``degree`` *distinct*
                input features (so not ``x[1] ** 2``, ``x[0] * x[2] ** 3``,
                etc.). Default is False.
            include_bias (bool, optional): If True, then include a bias column, the
                feature in which all polynomial powers are zero (i.e., a column of
                ones - acts as an intercept term in a linear model). Default is True.
            order (str in {'C', 'F'}, optional): Order of output array in the dense
                case. 'F' order is faster to compute, but may slow down subsequent
                estimators. Default is 'C'.

        Raises:
            ValueError: If degree is less than 1.
        """
        if degree == 0:
            raise ValueError(
                "degree must be at least 1, otherwise inverse cannot be " "computed"
            )
        super(Polynomial, self).__init__(
            degree=degree,
            interaction_only=interaction_only,
            include_bias=include_bias,
            order=order,
        )
        self.include_state = True

    def fit(self, x, y=None):
        """
        Compute number of output features.

        This method fits the `Polynomial` instance to the input data `x`. It calls the
        `fit` method of the superclass (`PolynomialFeatures` from
        `sklearn.preprocessing`), which computes the number of output features based
        on the degree of the polynomial and the interaction_only flag. It also sets
        `n_input_features_` and `n_output_features_` attributes. Then, it initializes
        `measurement_matrix_` as a zero matrix of size `n_input_features_` by
        `n_output_features_` and sets the main diagonal to 1, depending on the
        `include_bias` attribute. The input `y` is not used in this method; it is
        only included to maintain compatibility with the usual interface of `fit`
        methods in scikit-learn.

        Args:
            x (np.ndarray): The measurement data to be fit, with shape (n_samples,
                n_features).
            y (array-like, optional): Dummy input. Defaults to None.

        Returns:
            self: A fit instance of `Polynomial`.

        Raises:
            ValueError: If the input data is not valid.
        """
        x = validate_input(x)
        self.n_consumed_samples = 0

        y_poly_out = super(Polynomial, self).fit(x.real, y)

        self.measurement_matrix_ = np.zeros([x.shape[1], y_poly_out.n_output_features_])
        if self.include_bias:
            self.measurement_matrix_[:, 1 : 1 + x.shape[1]] = np.eye(x.shape[1])
        else:
            self.measurement_matrix_[:, : x.shape[1]] = np.eye(x.shape[1])

        return y_poly_out

    def transform(self, x):
        """
        Transforms the data to polynomial features.

        This method transforms the data `x` into polynomial features. It first checks if
        the fit method has been called by checking the `n_input_features_` attribute,
        then it validates the input `x`. If `x` is a CSR sparse matrix and the degree is
        less than 4, it uses a method based on "Leveraging Sparsity to Speed Up
        Polynomial Feature Expansions of CSR Matrices Using K-Simplex Numbers" by
        Andrew Nystrom and John Hughes. If `x` is a CSC sparse matrix and the degree
        is less than 4, it converts `x` to CSR, generates the polynomial features,
        then converts back to CSC. For dense arrays or CSC sparse matrix with a
        degree of 4 or more, it generates the polynomial features through a slower
        process.

        Args:
            x (array-like or CSR/CSC sparse matrix): The data to transform, row by row.
            The shape should be (n_samples, n_features). Prefer CSR over CSC for
            sparse input (for speed), but CSC is required if the degree is 4 or higher.

        Returns:
            xp (np.ndarray or CSR/CSC sparse matrix): The matrix of features, where
            n_output_features is the number of polynomial features generated from the
            combination of inputs. The shape is (n_samples, n_output_features).

        Raises:
            ValueError: If the input data is not valid or the shape of `x` does not
            match training shape.
        """
        check_is_fitted(self, "n_input_features_")

        x = check_array(x, order="F", dtype=FLOAT_DTYPES, accept_sparse=("csr", "csc"))

        n_samples, n_features = x.shape

        if n_features != self.n_input_features_:
            raise ValueError("x shape does not match training shape")

        if sparse.isspmatrix_csr(x):
            if self.degree > 3:
                return self.transform(x.tocsc()).tocsr()
            to_stack = []
            if self.include_bias:
                to_stack.append(np.ones(shape=(n_samples, 1), dtype=x.dtype))
            to_stack.append(x)
            for deg in range(2, self.degree + 1):
                xp_next = _csr_polynomial_expansion(
                    x.data,
                    x.indices,
                    x.indptr,
                    x.shape[1],
                    self.interaction_only,
                    deg,
                )
                if xp_next is None:
                    break
                to_stack.append(xp_next)
            xp = sparse.hstack(to_stack, format="csr")
        elif sparse.isspmatrix_csc(x) and self.degree < 4:
            return self.transform(x.tocsr()).tocsc()
        else:
            combinations = self._combinations(
                n_features,
                self.degree,
                self.interaction_only,
                self.include_bias,
            )
            if sparse.isspmatrix(x):
                columns = []
                for comb in combinations:
                    if comb:
                        out_col = 1
                        for col_idx in comb:
                            out_col = x[:, col_idx].multiply(out_col)
                        columns.append(out_col)
                    else:
                        bias = sparse.csc_matrix(np.ones((x.shape[0], 1)))
                        columns.append(bias)
                xp = sparse.hstack(columns, dtype=x.dtype).tocsc()
            else:
                xp = np.empty(
                    (n_samples, self.n_output_features_),
                    dtype=x.dtype,
                    order=self.order,
                )
                for i, comb in enumerate(combinations):
                    xp[:, i] = x[:, comb].prod(1)

        return xp

    @staticmethod
    def _combinations(n_features, degree, interaction_only, include_bias):
        """
        Generate combinations for polynomial features.

        This static method generates combinations of features for the polynomial
        transformation. The combinations depend on whether interaction_only is set
        and whether a bias term should be included.

        Args:
            n_features (int): The number of features.
            degree (int): The degree of the polynomial.
            interaction_only (bool): If True, only interaction features are produced.
            include_bias (bool): If True, a bias column is included.

        Returns:
            itertools.chain: An iterable over all combinations.
        """
        comb = combinations if interaction_only else combinations_w_r
        start = int(not include_bias)
        return chain.from_iterable(
            comb(range(n_features), i) for i in range(start, degree + 1)
        )

    @property
    def powers_(self):
        """
        Get the exponent for each of the inputs in the output.

        This property method returns the exponents for each input feature in the
        polynomial output. It first checks whether the model has been fitted, then
        uses the `_combinations` method to get the combinations of features, and
        finally calculates the exponents for each input feature.

        Returns:
            np.ndarray: A 2D array where each row represents a feature and each
            column represents an output of the polynomial transformation. The
            values are the exponents of the input features.

        Raises:
            NotFittedError: If the model has not been fitted.
        """
        check_is_fitted(self)

        combinations = self._combinations(
            n_features=self.n_input_features_,
            degree=self.degree,
            interaction_only=self.interaction_only,
            include_bias=self.include_bias,
        )
        return np.vstack(
            [np.bincount(c, minlength=self.n_input_features_) for c in combinations]
        )



"""moduel for time-delay observables"""
from __future__ import annotations

import numpy as np
from numpy import arange
from numpy import empty
from sklearn.utils.validation import check_is_fitted

from ..common import validate_input
from ._base import BaseObservables


class TimeDelay(BaseObservables):
    """
    A class for creating time-delay observables. These observables are formed by
    taking time-lagged measurements of state variables and interpreting them as new
    state variables.

    The two state variables :math:`[x(t), y(t)]` could be supplemented with two
    time-delays each, yielding a new set of observables:

    .. math::
        [x(t), y(t), x(t-\\Delta$ t), y(t-\\Delta t),
        x(t-2\\Delta t), y(t - 2\\Delta t)]

    This example corresponds to taking :code:`delay =` :math:`\\Delta t` and
    :code:`n_delays = 2`.

    Note that when transforming data the first :code:`delay * n_delays` rows/samples
    are dropped as there is insufficient time history to form time-delays for them.

    For more information, see the following references:

        Brunton, Steven L., et al.
        "Chaos as an intermittently forced linear system."
        Nature communications 8.1 (2017): 1-9.

        Susuki, Yoshihiko, and Igor Mezić.
        "A prony approximation of Koopman mode decomposition."
        2015 54th IEEE Conference on Decision and Control (CDC). IEEE, 2015.

        Arbabi, Hassan, and Igor Mezic.
        "Ergodic theory, dynamic mode decomposition, and computation
        of spectral properties of the Koopman operator."
        SIAM Journal on Applied Dynamical Systems 16.4 (2017): 2096-2126.

    Args:
        delay (int, optional): The length of each delay. Defaults to 1.
        n_delays (int, optional): The number of delays to compute for each
            variable. Defaults to 2.

    Attributes:
        include_state (bool): If True, includes the system state.
        delay (int): The length of each delay.
        n_delays (int): The number of delays to compute for each variable.
        _n_consumed_samples (int): Number of samples consumed when :code:`transform`
            is called,i.e. :code:`n_delays * delay`.
    """

    def __init__(self, delay=1, n_delays=2):
        """
        Initialize the TimeDelay class with given parameters.

        Args:
            delay (int, optional): The length of each delay. Defaults to 1. Or
                we say this is the "stride of delay".
            n_delays (int, optional): The number of delays to compute for each
                variable. Defaults to 2.

        Raises:
            ValueError: If delay or n_delays are negative.
        """
        super(TimeDelay, self).__init__()
        if delay < 0:
            raise ValueError("delay must be a nonnegative int")
        if n_delays < 0:
            raise ValueError("n_delays must be a nonnegative int")

        self.include_state = True
        self.delay = int(delay)
        self.n_delays = int(n_delays)
        self._n_consumed_samples = self.delay * self.n_delays

    def fit(self, x, y=None):
        """
        Fit the model to measurement data.

        Args:
            x (array-like): The input data, shape (n_samples, n_input_features).
            y (None): Dummy parameter for sklearn compatibility.

        Returns:
            TimeDelay: The fitted instance.
        """

        x = validate_input(x)
        n_samples, n_features = x.shape

        self.n_input_features_ = n_features
        self.n_output_features_ = n_features * (1 + self.n_delays)

        self.measurement_matrix_ = np.zeros(
            (self.n_input_features_, self.n_output_features_)
        )
        self.measurement_matrix_[
            : self.n_input_features_, : self.n_input_features_
        ] = np.eye(self.n_input_features_)

        return self

    def transform(self, x):
        """
        Add time-delay features to the data, dropping the first :code:`delay -
        n_delays` samples.

        Args:
            x (array-like): The input data, shape (n_samples, n_input_features).
                It is assumed that rows correspond to examples that are equi-spaced
                in time and are in sequential order.

        Returns:
            y (array-like): The transformed data, shape (n_samples - delay * n_delays,
            n_output_features).
        """

        check_is_fitted(self, "n_input_features_")
        x = validate_input(x)

        if x.shape[1] != self.n_input_features_:
            raise ValueError(
                "Wrong number of input features. "
                f"Expected x.shape[1] = {self.n_input_features_}; "
                f"instead x.shape[1] = {x.shape[1]}."
            )

        self._n_consumed_samples = self.delay * self.n_delays
        if len(x) < self._n_consumed_samples + 1:
            raise ValueError(
                "x has too few rows. To compute time-delay features with "
                f"delay = {self.delay} and n_delays = {self.n_delays} "
                f"x must have at least {self._n_consumed_samples + 1} rows."
            )

        y = empty(
            (x.shape[0] - self._n_consumed_samples, self.n_output_features_),
            dtype=x.dtype,
        )
        y[:, : self.n_input_features_] = x[self._n_consumed_samples :]

        for i in range(self._n_consumed_samples, x.shape[0]):
            y[i - self._n_consumed_samples, self.n_input_features_ :] = x[
                self._delay_inds(i), :
            ].flatten()

        return y

    def get_feature_names(self, input_features=None):
        """
        Get the names of the output features.

        Args:
            input_features (list of str, optional): Names for input features.
                If None, defaults to "x0", "x1", ... ,"xn_input_features".

        Returns:
            list of str: Names of the output features.
        """
        check_is_fitted(self, "n_input_features_")
        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )

        output_features = [f"{xi}(t)" for xi in input_features]
        output_features.extend(
            [
                f"{xi}(t-{i * self.delay}dt)"
                for i in range(1, self.n_delays + 1)
                for xi in input_features
            ]
        )

        return output_features

    def _delay_inds(self, index):
        """
        Private method to get the indices for the delayed data.

        Args:
            index (int): The index from which to calculate the delay indices.

        Returns:
            array-like: The delay indices.
        """
        return index - self.delay * arange(1, self.n_delays + 1)

    @property
    def n_consumed_samples(self):
        """
        The number of samples that are consumed as "initial conditions" for
        other samples, i.e., the number of samples for which time delays cannot
        be computed.

        Returns:
            int: The number of consumed samples.
        """
        return self._n_consumed_samples



"""module for 1D KS equation"""
from __future__ import annotations

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from scipy.fft import fft
from scipy.fft import fftfreq
from scipy.fft import ifft


class ks:
    """
    Solving 1D KS equation

    u_t = -u*u_x + u_{xx} + nu*u_{xxxx}

    Periodic B.C. between 0 and 2*pi. This PDE is solved
    using spectral methods.
    """

    def __init__(self, n, x, nu, dt, M=16):
        self.n_states = n
        self.dt = dt
        self.x = x
        dk = 1
        k = fftfreq(self.n_states, 1.0 / self.n_states) * dk
        k[n // 2] = 0.0
        L = k**2 - nu * k**4
        self.E = np.exp(self.dt * L)
        self.E2 = np.exp(self.dt * L / 2.0)
        # self.M = M
        r = np.exp(1j * np.pi * (np.arange(1, M + 1) - 0.5) / M)
        r = r.reshape(1, -1)
        r_on_circle = np.repeat(r, n, axis=0)
        LR = self.dt * L
        LR = LR.reshape(-1, 1)
        LR = LR.astype("complex")
        LR = np.repeat(LR, M, axis=1)
        LR += r_on_circle
        self.g = -0.5j * k

        self.Q = self.dt * np.real(np.mean((np.exp(LR / 2.0) - 1) / LR, axis=1))
        self.f1 = self.dt * np.real(
            np.mean(
                (-4.0 - LR + np.exp(LR) * (4.0 - 3.0 * LR + LR**2)) / LR**3, axis=1
            )
        )
        self.f2 = self.dt * np.real(
            np.mean((2.0 + LR + np.exp(LR) * (-2.0 + LR)) / LR**3, axis=1)
        )
        self.f3 = self.dt * np.real(
            np.mean(
                (-4.0 - 3.0 * LR - LR**2 + np.exp(LR) * (4.0 - LR)) / LR**3, axis=1
            )
        )

    @staticmethod
    def compute_u2k_zeropad_dealiased(uk_):
        # three over two law
        N = uk_.size
        # map uk to uk_fine
        uk_fine = (
            np.hstack((uk_[0 : int(N / 2)], np.zeros((int(N / 2))), uk_[int(-N / 2) :]))
            * 3.0
            / 2.0
        )
        # convert uk_fine to physical space
        u_fine = np.real(ifft(uk_fine))
        # compute square
        u2_fine = np.square(u_fine)
        # compute fft on u2_fine
        u2k_fine = fft(u2_fine)
        # convert u2k_fine to u2k
        u2k = np.hstack((u2k_fine[0 : int(N / 2)], u2k_fine[int(-N / 2) :])) / 3.0 * 2.0
        return u2k

    def sys(self, t, x, u):
        raise NotImplementedError

    def simulate(self, x0, n_int, n_sample):
        xk = fft(x0)
        u = np.zeros((n_int, 1))
        X = np.zeros((n_int // n_sample, self.n_states))
        t = 0
        j = 0
        t_list = []
        for step in range(n_int):
            t += self.dt
            Nv = self.g * self.compute_u2k_zeropad_dealiased(xk)
            a = self.E2 * xk + self.Q * Nv
            Na = self.g * self.compute_u2k_zeropad_dealiased(a)
            b = self.E2 * xk + self.Q * Na
            Nb = self.g * self.compute_u2k_zeropad_dealiased(b)
            c = self.E2 * a + self.Q * (2.0 * Nb - Nv)
            Nc = self.g * self.compute_u2k_zeropad_dealiased(c)
            xk = self.E * xk + Nv * self.f1 + 2.0 * (Na + Nb) * self.f2 + Nc * self.f3

            if (step + 1) % n_sample == 0:
                y = np.real(ifft(xk)) + self.dt * u[j]
                X[j, :] = y
                j += 1
                t_list.append(t)
                xk = fft(y)

        return X, np.array(t_list)

    def collect_data_continuous(self, x0):
        raise NotImplementedError

    def collect_one_step_data_discrete(self, x0):
        """
        collect training data pairs - discrete sense.

        given x0, with shape (n_dim, n_traj), the function
        returns system state x1 after self.dt with shape
        (n_dim, n_traj)
        """
        n_traj = x0.shape[0]
        X = x0
        Y = []
        for i in range(n_traj):
            y, _ = self.simulate(x0[i], n_int=1, n_sample=1)
            Y.append(y)
        Y = np.vstack(Y)
        return X, Y

    def collect_one_trajectory_data(self, x0, n_int, n_sample):
        x = x0
        y, _ = self.simulate(x, n_int, n_sample)
        return y

    def visualize_data(self, x, t, X):
        plt.figure(figsize=(6, 6))
        ax = plt.axes(projection=Axes3D.name)
        for i in range(X.shape[0]):
            ax.plot(x, X[i], zs=t[i], zdir="t", label="time = " + str(i * self.dt))
        ax.view_init(elev=35.0, azim=-65, vertical_axis="y")
        ax.set(ylabel=r"$u(x,t)$", xlabel=r"$x$", zlabel=r"time $t$")
        plt.title("1D K-S equation")
        plt.show()

    def visualize_state_space(self, X):
        u, s, vt = np.linalg.svd(X, full_matrices=False)
        plt.figure(figsize=(6, 6))
        plt.semilogy(s)
        plt.xlabel("number of SVD terms")
        plt.ylabel("singular values")
        plt.title("PCA singular value decays")
        plt.show()

        # this is a pde problem so the number of snapshots are smaller than dof
        pca_1, pca_2, pca_3 = u[:, 0], u[:, 1], u[:, 2]
        plt.figure(figsize=(6, 6))
        ax = plt.axes(projection=Axes3D.name)
        ax.plot3D(pca_1, pca_2, pca_3, "k-o")
        ax.set(xlabel="pc1", ylabel="pc2", zlabel="pc3")
        plt.title("PCA visualization")
        plt.show()


if __name__ == "__main__":
    n = 256
    x = np.linspace(0, 2.0 * np.pi, n, endpoint=False)
    u0 = np.sin(x)
    nu = 0.01
    n_int = 1000
    n_snapshot = 500
    dt = 4.0 / n_int
    n_sample = n_int // n_snapshot

    model = ks(n, x, nu=nu, dt=dt)
    X, t = model.simulate(u0, n_int, n_sample)
    print(X.shape)
    model.visualize_data(x, t, X)

    # usage: visualize the data in state space
    model.visualize_state_space(X)

    # usage: collect discrete data pair
    x0_array = np.vstack([u0, u0, u0])
    X, Y = model.collect_one_step_data_discrete(x0_array)

    print(X.shape)
    print(Y.shape)

    # usage: collect one trajectory data
    X = model.collect_one_trajectory_data(u0, n_int, n_sample)
    print(X.shape)



"""module for base class of regressor"""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod

from sklearn.base import BaseEstimator


class BaseRegressor(BaseEstimator, ABC):
    """
    Base class for PyKoopman regressors.

    This class provides a wrapper for regressors used in the PyKoopman package.
    It's designed to be used with any regressor object that implements `fit`
    and `predict` methods following the `sklearn.base.BaseEstimator` interface.

    Note: This is an abstract base class, and should not be instantiated directly.
    Instead, a subclass should be created that implements the required abstract methods.

    Args:
        regressor (BaseEstimator): A regressor object implementing `fit` and `predict`
        methods.

    Attributes:
        regressor (BaseEstimator): The regressor object passed during initialization.

    Abstract methods:
        coef_ : Should return the coefficients of the regression model.

        state_matrix_ : Should return the state matrix of the dynamic system.

        eigenvectors_ : Should return the eigenvectors of the system.

        eigenvalues_ : Should return the eigenvalues of the system.

        _compute_phi(x_col) : Should compute and return the phi function on given data.

        _compute_psi(x_col) : Should compute and return the psi function on given data.

        ur : Should return the u_r of the system.

        unnormalized_modes : Should return the unnormalized modes of the system.
    """

    def __init__(self, regressor):
        # check .fit
        if not hasattr(regressor, "fit") or not callable(getattr(regressor, "fit")):
            raise AttributeError("regressor does not have a callable fit method")
        # check .predict
        if not hasattr(regressor, "predict") or not callable(
            getattr(regressor, "predict")
        ):
            raise AttributeError("regressor does not have a callable predict method")
        self.regressor = regressor

    def fit(self, x, y=None):
        raise NotImplementedError

    def predict(self, x):
        raise NotImplementedError

    @abstractmethod
    def coef_(self):
        pass

    @abstractmethod
    def state_matrix_(self):
        pass

    @abstractmethod
    def eigenvectors_(self):
        pass

    @abstractmethod
    def eigenvalues_(self):
        pass

    @abstractmethod
    def _compute_phi(self, x_col):
        pass

    @abstractmethod
    def _compute_psi(self, x_col):
        pass

    @abstractmethod
    def ur(self):
        pass

    @abstractmethod
    def unnormalized_modes(self):
        pass



"""module for havok"""
from __future__ import annotations

from warnings import warn

import numpy as np
from matplotlib import pyplot as plt
from optht import optht
from scipy.signal import lsim
from scipy.signal import lti
from sklearn.utils.validation import check_is_fitted

from ..common import drop_nan_rows
from ..differentiation._derivative import Derivative
from ._base import BaseRegressor


class HAVOK(BaseRegressor):
    """
    HAVOK (Hankel Alternative View of Koopman) regressor.

    Aims to determine the system matrices A, B that satisfy d/dt v = Av + Bu,
    where v is the vector of the leading delay coordinates and u is a low-energy
    delay coordinate acting as forcing. A and B are the unknown system and control
    matrices, respectively. The delay coordinates are obtained by computing the
    SVD from a Hankel matrix.

    The objective function, \\|dV-AV-BU\\|_F, is minimized using least-squares
    regression.

    See the following reference for more details:
        Brunton, S.L., Brunton, B.W., Proctor, J.L., Kaiser, E. & Kutz, J.N.
        "Chaos as an intermittently forced linear system."
        Nature Communications, Vol. 8(19), 2017.
        <https://www.nature.com/articles/s41467-017-00030-8>

    Parameters:
        svd_rank (int, optional):
            Rank of the SVD used for model reduction. Defaults to None.
        differentiator (Derivative, optional):
            Differentiation method to compute the time derivative. Defaults to
            Derivative(kind="finite_difference", k=1).
        plot_sv (bool, optional):
            Whether to plot the singular values. Defaults to False.

    Attributes:
        coef_ (array):
            Weight vectors of the regression problem. Corresponds to either [A] or
            [A,B].
        state_matrix_ (array):
            Identified state transition matrix A of the underlying system.
        control_matrix_ (array):
            Identified control matrix B of the underlying system.
        projection_matrix_ (array):
            Projection matrix into low-dimensional subspace of shape (n_input_features
            +n_control_features, svd_rank).
        projection_matrix_output_ (array):
            Projection matrix into low-dimensional subspace of shape (n_input_features
            +n_control_features, svd_output_rank).
    """

    def __init__(
        self,
        svd_rank=None,
        differentiator=Derivative(kind="finite_difference", k=1),
        plot_sv=False,
    ):
        """
        Initialize the HAVOK regressor.

        Args:
            svd_rank (int, optional):
                Rank of the SVD used for model reduction. Defaults to None.
            differentiator (Derivative, optional):
                Differentiation method to compute the time derivative. Defaults to
                Derivative(kind="finite_difference", k=1).
            plot_sv (bool, optional):
                Whether to plot the singular values. Defaults to False.
        """
        self.svd_rank = svd_rank
        self.differentiator = differentiator
        self.plot_sv = plot_sv

    def fit(self, x, y=None, dt=None):
        """
        Fit the HAVOK regressor to the given data.

        Args:
            x (numpy.ndarray):
                Measurement data to be fit.
            y (not used):
                Time-shifted measurement data to be fit. Ignored.
            dt (scalar):
                Discrete time-step.

        Returns:
            self: Fitted HAVOK instance.
        """

        if y is not None:
            warn("havok regressor does not require the y argument when fitting.")

        if dt is None:
            raise ValueError("havok regressor requires a timestep dt when fitting.")

        self.dt_ = dt
        self.n_samples_, self.n_input_features_ = x.shape
        self.n_control_features_ = 1

        # Create time vector
        t = np.arange(0, self.dt_ * self.n_samples_, self.dt_)

        # SVD to calculate intrinsic observables
        U, s, Vh = np.linalg.svd(x.T, full_matrices=False)

        if self.plot_sv:
            plt.figure()
            plt.semilogy(s)
            plt.xlabel("number of terms")
            plt.ylabel("singular values")
            plt.show()

        # calculate rank using optimal hard threshold by Gavish & Donoho
        if self.svd_rank is None:
            self.svd_rank = optht(x, sv=s, sigma=None)
        Vrh = Vh[: self.svd_rank, :]
        Vr = Vrh.T
        Ur = U[:, : self.svd_rank]
        sr = s[: self.svd_rank]

        # calculate time derivative dxdt of only the first rank-1 & normalize
        dVr = self.differentiator(Vr[:, :-1], t)
        # this line actually makes vh and dvh transposed
        dVr, t, V = drop_nan_rows(dVr, t, Vh.T)

        # regression on intrinsic variables v
        # xi = np.zeros((self.svd_rank - 1, self.svd_rank))
        # for i in range(self.svd_rank - 1):
        #     # here, we use rank terms in V to fit the rank-1 terms dV/dt
        #     # we perform column wise
        #     xi[i, :] = np.linalg.lstsq(Vr, dVr[:, i], rcond=None)[0]

        xi = np.linalg.lstsq(Vr, dVr, rcond=None)[0].T
        assert xi.shape == (self.svd_rank - 1, self.svd_rank)

        self.forcing_signal = Vr[:, -1]
        self._state_matrix_ = xi[:, :-1]
        self._control_matrix_ = xi[:, -1].reshape(-1, 1)

        self.svals = s
        self._ur = Ur[:, :-1] @ np.diag(sr[:-1])
        self._coef_ = np.hstack([self.state_matrix_, self.control_matrix_])

        eigenvalues_, self._eigenvectors_ = np.linalg.eig(self.state_matrix_)
        # because we fit the model in continuous time,
        # so we need to convert to discrete time
        self._eigenvalues_ = np.exp(eigenvalues_ * dt)

        self._unnormalized_modes = self._ur @ self.eigenvectors_
        self._tmp_compute_psi = np.linalg.inv(self.eigenvectors_) @ self._ur.T

        # self.C = np.linalg.multi_dot(
        #     [
        #         np.linalg.inv(self.eigenvectors_),
        #         np.diag(np.reciprocal(s[: self.svd_rank - 1])),
        #         U[:, : self.svd_rank - 1].T,
        #     ]
        # )

    def predict(self, x, u, t):
        """
        Predict the output based on the input data.

        Args:
            x (numpy.ndarray):
                Measurement data upon which to base prediction.
            u (numpy.ndarray):
                Time series of external actuation/control, which is sampled at time
                instances in `t`.
            t (numpy.ndarray):
                Time vector. Instances at which the solution vector shall be provided.
                Note: The time vector must start at 0.

        Returns:
            y (numpy.ndarray):
                Prediction of `x` at the time instances provided in `t`.
        """
        # if t[0] != 0:
        #    raise ValueError("the time vector must start at 0.")

        check_is_fitted(self, "coef_")
        y0 = (
            # np.linalg.inv(np.diag(self.svals[: self.svd_rank - 1]))
            # @
            np.linalg.pinv(self._ur)
            @ x.T
        )
        sys = lti(
            self.state_matrix_,
            self.control_matrix_,
            self._ur,
            np.zeros((self.n_input_features_, self.n_control_features_)),
        )
        tout, ypred, xpred = lsim(sys, U=u, T=t, X0=y0.T)
        return ypred

    def _compute_phi(self, x_col):
        """
        Compute the feature vector `phi(x)` given `x`.

        Args:
            x_col (numpy.ndarray):
                Input data `x` for computing `phi(x)`.

        Returns:
            phi (numpy.ndarray):
                Value of `phi(x)`.

        """
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        phi = self._ur.T @ x_col
        return phi

    def _compute_psi(self, x_col):
        """
        Compute the feature vector `psi(x)` given `x`.

        Args:
            x_col (numpy.ndarray):
                Input data `x` for computing `psi(x)`.

        Returns:
            psi (numpy.ndarray):
                Value of `psi(x)`.

        """
        # compute psi - one column if x is a row
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        psi = self._tmp_compute_psi @ x_col
        return psi

    @property
    def coef_(self):
        """
        Get the weight vectors of the regression problem.

        Returns:
            coef (numpy.ndarray):
                Weight vectors of the regression problem. Corresponds to either [A]
                or [A,B].
        """
        check_is_fitted(self, "_coef_")
        return self._coef_

    @property
    def state_matrix_(self):
        """
        Get the identified state transition matrix A of the underlying system.

        Returns:
            state_matrix (numpy.ndarray):
                Identified state transition matrix A.
        """
        check_is_fitted(self, "_state_matrix_")
        return self._state_matrix_

    @property
    def control_matrix_(self):
        """
        Get the identified control matrix B of the underlying system.

        Returns:
            control_matrix (numpy.ndarray):
                Identified control matrix B.
        """
        check_is_fitted(self, "_control_matrix_")
        return self._control_matrix_

    @property
    def eigenvectors_(self):
        """
        Get the identified eigenvectors of the state matrix A.

        Returns:
            eigenvectors (numpy.ndarray):
                Identified eigenvectors of the state matrix A.
        """
        check_is_fitted(self, "_eigenvectors_")
        return self._eigenvectors_

    @property
    def eigenvalues_(self):
        """
        Get the identified eigenvalues of the state matrix A.

        Returns:
            eigenvalues (numpy.ndarray):
                Identified eigenvalues of the state matrix A.
        """
        check_is_fitted(self, "_eigenvalues_")
        return self._eigenvalues_

    @property
    def unnormalized_modes(self):
        """
        Get the identified unnormalized modes.

        Returns:
            unnormalized_modes (numpy.ndarray):
                Identified unnormalized modes.
        """
        check_is_fitted(self, "_unnormalized_modes")
        return self._unnormalized_modes

    @property
    def ur(self):
        """
        Get the matrix UR.

        Returns:
            ur (numpy.ndarray):
                Matrix UR.
        """
        check_is_fitted(self, "_ur")
        return self._ur



"""module for kernel dmd"""
from __future__ import annotations

from warnings import warn

import numpy as np
from pydmd.dmdbase import DMDTimeDict
from pydmd.utils import compute_svd
from pydmd.utils import compute_tlsq
from scipy.linalg import sqrtm
from sklearn.gaussian_process.kernels import Kernel
from sklearn.gaussian_process.kernels import RBF
from sklearn.utils.validation import check_is_fitted

from ._base import BaseRegressor


class KDMD(BaseRegressor):
    """
    Kernel Dynamic Mode Decomposition.

    See the following reference for more details:

    `Williams, M. O., Rowley, C. W., & Kevrekidis, I. G. (2014).
    "A kernel-based approach to data-driven Koopman spectral analysis."
    arXiv preprint arXiv:1411.2260. <https://arxiv.org/pdf/1411.2260.pdf>`

    Args:
        svd_rank (int): The rank for the truncation. If 0, the method computes
            the optimal rank and uses it for truncation. If positive integer,
            the method uses the argument for the truncation. If float between 0
            and 1, the rank is the number of the biggest singular values that
            are needed to reach the 'energy' specified by `svd_rank`. If -1,
            the method does not compute truncation. Default is 0.
        tlsq_rank (int): The rank for the truncation. If 0, the method does not
            compute any noise reduction. If positive number, the method uses the
            argument for the SVD truncation used in the TLSQ method.
        forward_backward (bool): If True, the low-rank operator is computed
            like in fbDMD (reference: https://arxiv.org/abs/1507.02264).
            Default is False.
        tikhonov_regularization (bool or None): Tikhonov parameter for the
            regularization. If None, no regularization is applied. If float,
            it is used as the λ Tikhonov parameter.
        kernel (sklearn.gaussian_process.Kernel): An instance of kernel from sklearn.

    Attributes:
        svd_rank (int): The rank for the truncation.
        tlsq_rank (int): The rank for the truncation.
        forward_backward (bool): If True, the low-rank operator is computed
            like in fbDMD (reference: https://arxiv.org/abs/1507.02264).
        tikhonov_regularization (bool or None): Tikhonov parameter for the
            regularization.
        kernel (sklearn.gaussian_process.Kernel): An instance of kernel from sklearn.
        n_samples_ (int): Number of samples in KDMD.
        n_input_features_ (int): Dimension of input features, i.e., the dimension
            of each sample.
        _snapshots (numpy.ndarray): Column-wise data matrix of shape
            (n_input_features_, n_samples_).
        _snapshots_shape (tuple): Shape of column-wise data matrix.
        _X (numpy.ndarray): Training features column-wise arranged, needed for
            prediction. Shape is (n_input_features_, n_samples).
        _Y (numpy.ndarray): Training target, column-wise arranged. Shape is
            (n_input_features_, n_samples).
        _coef_ (numpy.ndarray): Reduced Koopman state transition matrix of shape
            (svd_rank, svd_rank).
        _eigenvalues_ (numpy.ndarray): Koopman lambda of shape (svd_rank,).
        _eigenvectors_ (numpy.ndarray): Koopman eigenvectors of shape
            (svd_rank, svd_rank).
        _unnormalized_modes (numpy.ndarray): Koopman V of shape
            (svd_rank, n_input_features_).
        _state_matrix_ (numpy.ndarray): Reduced Koopman state transition matrix
            of shape (svd_rank, svd_rank).
        self.C (numpy.ndarray): Linear matrix that maps kernel product features
            to eigenfunctions of shape (svd_rank, n_samples_).
    """

    def __init__(
        self,
        svd_rank=1.0,  # 1.0 means keeping all ranks
        tlsq_rank=0,
        forward_backward=False,
        tikhonov_regularization=None,
        kernel=RBF(),
    ):
        """
        Kernel Dynamic Mode Decomposition.

        Args:
            svd_rank (int, optional): The rank for the truncation.
                If set to 0, the method computes the optimal rank
                and uses it for truncation. If set to a positive integer,
                the method uses the specified rank for truncation.
                If set to a float between 0 and 1, the rank is determined
                based on the specified energy level. If set to -1, no
                truncation is performed. Default is 1.0.
            tlsq_rank (int, optional): The rank for the truncation used
                in the total least squares preprocessing. If set to 0,
                no noise reduction is performed. If set to a positive integer,
                the method uses the specified rank for the SVD truncation
                in the TLSQ method. Default is 0.
            forward_backward (bool, optional): Whether to compute the
                low-rank operator using the forward-backward method similar
                to fbDMD. If set to True, the low-rank operator is computed
                with forward-backward DMD. If set to False, standard DMD is used.
                Default is False.
            tikhonov_regularization (float or None, optional): Tikhonov
                regularization parameter for regularization. If set to None,
                no regularization is applied. If set to a float, it is used
                as the regularization parameter. Default is None.
            kernel (Kernel, optional): An instance of the kernel class from
                sklearn.gaussian_process. Default is RBF().
        """
        self.svd_rank = svd_rank
        self.tlsq_rank = tlsq_rank
        self.forward_backward = forward_backward
        self.tikhonov_regularization = tikhonov_regularization
        self.kernel = kernel

        if not isinstance(self.kernel, Kernel):
            raise ValueError(
                "kernel must be a subclass of sklearn.gaussian_process.kernel"
            )

    def fit(self, x, y=None, dt=1):
        """
        Fits the KDMD model to the provided training data.

        Args:
            x: numpy.ndarray, shape (n_samples, n_features)
                Measurement data input.

            y: numpy.ndarray, shape (n_samples, n_features), optional
                Measurement data output to be fitted. Defaults to None.

            dt: float, optional
                Time interval between `x` and `y`. Defaults to 1.

        Returns:
            KDMD:
                The fitted KDMD instance.
        """

        # if y is not None:
        #    warn("pydmd regressors do not require the y argument when fitting.")
        self.n_samples_, self.n_input_features_ = x.shape
        n_samples = self.n_samples_
        if y is None:
            self._snapshots, self._snapshots_shape = _col_major_2darray(x.T)
            # self._snapshots.shape[1]
            X = self._snapshots[:, :-1]
            Y = self._snapshots[:, 1:]
        else:
            # if we have pairs of data
            X = x.T
            Y = y.T

        # total least square preprocessing on X and Y - features, samples
        self._X, self._Y = compute_tlsq(X, Y, self.tlsq_rank)

        # compute KDMD operators, lamda, and koopman V
        # note that this method is built by considering row-wise collected data
        [
            self._coef_,
            self._eigenvalues_,
            self._eigenvectors_,
            self._unnormalized_modes,
        ] = self._regressor_compute_kdmdoperator(self._X.T, self._Y.T)

        # Default timesteps
        self._set_initial_time_dictionary({"t0": 0, "tend": n_samples - 1, "dt": 1})

        # _coef_ as the transpose
        # self._coef_ = self._regressor_atilde.T

        return self

    def predict(self, x):
        """
        Predicts the future states based on the given input data.

        Args:
            x: numpy.ndarray, shape (n_samples, n_features)
                Measurement data upon which to base the prediction.

        Returns:
            numpy.ndarray, shape (n_samples, n_features)
                Prediction of the future states.
        """

        check_is_fitted(self, "coef_")

        phi = self._compute_psi(x_col=x.T)
        phi_next = np.diag(self.eigenvalues_) @ phi
        x_next_T = self._unnormalized_modes @ phi_next
        return np.real(x_next_T).T

    def _compute_phi(self, x_col):
        """
        Computes the phi(x) given x.

        Args:
            x_col: numpy.ndarray, shape (n_samples, n_features)
                Measurement data upon which to compute phi values.

        Returns:
            numpy.ndarray, shape (n_samples, n_input_features_)
                Value of phi at x.
        """
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        psi = self._compute_psi(x_col)
        phi = np.real(self.eigenvectors_ @ psi)
        return phi

    def _compute_psi(self, x_col):
        """
        Computes the psi(x) given x.

        Args:
            x_col: numpy.ndarray, shape (n_samples, n_features)
                Measurement data upon which to compute psi values.

        Returns:
            numpy.ndarray, shape (n_samples, n_input_features_)
                Value of psi at x.
        """
        # compute psi - one column if x is a row
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        return self._tmp_compute_psi_kdmd @ self.kernel(self._X.T, x_col.T)

    @property
    def coef_(self):
        """
        Getter property for the coef_ attribute.

        Returns:
            numpy.ndarray, shape (svd_rank, svd_rank)
                Reduced Koopman state transition matrix.
        """
        check_is_fitted(self, "_coef_")
        return self._coef_

    @property
    def state_matrix_(self):
        """
        Getter property for the state_matrix_ attribute.

        Returns:
            numpy.ndarray, shape (svd_rank, svd_rank)
                Reduced Koopman state transition matrix.
        """
        check_is_fitted(self, "_state_matrix_")
        return self._state_matrix_

    @property
    def eigenvalues_(self):
        """
        Getter property for the eigenvalues_ attribute.

        Returns:
            numpy.ndarray, shape (svd_rank,)
                Koopman eigenvalues.
        """
        check_is_fitted(self, "_eigenvalues_")
        return self._eigenvalues_

    @property
    def eigenvectors_(self):
        """
        Getter property for the eigenvectors_ attribute.

        Returns:
            numpy.ndarray, shape (svd_rank, svd_rank)
                Koopman eigenvectors.
        """
        check_is_fitted(self, "_eigenvectors_")
        return self._eigenvectors_

    @property
    def unnormalized_modes(self):
        """
        Getter property for the unnormalized_modes attribute.

        Returns:
            numpy.ndarray, shape (svd_rank, n_input_features_)
                Koopman unnormalized modes.
        """
        check_is_fitted(self, "_unnormalized_modes")
        return self._unnormalized_modes

    @property
    def ur(self):
        """
        Getter property for the ur attribute.

        Returns:
            numpy.ndarray, shape (n_samples_, n_input_features_)
                Linear matrix that maps kernel product features to eigenfunctions.
        """
        check_is_fitted(self, "_ur")
        return self._ur

    def _regressor_compute_kdmdoperator(self, X, Y):
        """
        Computes the KDMD operator given input data X and target data Y.

        Args:
            X: numpy.ndarray, shape (n_samples_, n_input_features_)
                Training data input.
            Y: numpy.ndarray, shape (n_samples_, n_input_features_)
                Training data target output.

        Returns:
            list
                A list containing the following elements:
                - koopman_matrix: numpy.ndarray, shape (svd_rank, svd_rank)
                    Reduced Koopman state transition matrix.
                - koopman_eigvals: numpy.ndarray, shape (svd_rank,)
                    Koopman eigenvalues.
                - koopman_eigenvectors: numpy.ndarray, shape (svd_rank, svd_rank)
                    Koopman eigenvectors.
                - unnormalized_modes: numpy.ndarray, shape (svd_rank, n_input_features_)
                    Koopman unnormalized modes.
        """
        # compute kernel K(X,X)
        # since sklearn kernel function takes rowwise collected data.
        KXX = self.kernel(X, X)
        KYX = self.kernel(Y, X)

        # compute eig of PD matrix, so it is SVD
        U, s2, _ = compute_svd(KXX, self.svd_rank)
        s = np.sqrt(s2)
        # remember that we need sigma, but svd or eig only gives you the s^2

        # optional compute tiknoiv reg
        if self.tikhonov_regularization is not None:
            s = (
                s**2 + self.tikhonov_regularization * np.linalg.norm(X)
            ) * np.reciprocal(s)

        koopman_matrix = (
            np.diag(np.reciprocal(s))
            @ U.T.conj()
            @ KYX.T
            @ U
            @ np.diag(np.reciprocal(s))
        )

        # optional compute fb
        if self.forward_backward:
            KYY = self.kernel(Y, Y)
            KXY = KYX.T
            bU, bs2, _ = compute_svd(KYY, self.svd_rank)
            bs = np.sqrt(bs2)
            if self.tikhonov_regularization is not None:
                bs = (
                    bs**2 + self.tikhonov_regularization * np.linalg.norm(Y)
                ) * np.reciprocal(bs)

            atilde_back = (
                np.diag(np.reciprocal(bs))
                @ bU.T.conj()
                @ KXY.T
                @ bU
                @ np.diag(np.reciprocal(bs))
            )
            koopman_matrix = sqrtm(koopman_matrix @ np.linalg.inv(atilde_back))

        # self._regressor_atilde = atilde
        self._state_matrix_ = koopman_matrix

        # compute eigenquantities
        koopman_eigvals, koopman_eigenvectors = np.linalg.eig(koopman_matrix)

        # compute unnormalized V
        BV = np.linalg.lstsq(U @ np.diag(s), X, rcond=None)[0].T
        unnormalized_modes = BV @ koopman_eigenvectors

        # compute psi
        self._ur = BV  # U @ np.diag(s)
        self._tmp_compute_psi_kdmd = (
            np.linalg.inv(koopman_eigenvectors) @ np.diag(np.reciprocal(s)) @ U.T
        )

        return [
            koopman_matrix,
            koopman_eigvals,
            koopman_eigenvectors,
            unnormalized_modes,
        ]

    def _set_initial_time_dictionary(self, time_dict):
        """
        Sets the initial time dictionary.

        Args:
            time_dict: dict
                Dictionary containing the time information with keys 't0', 'tend',
                and 'dt'.
        """
        if not ("t0" in time_dict and "tend" in time_dict and "dt" in time_dict):
            raise ValueError('time_dict must contain the keys "t0", "tend" and "dt".')
        if len(time_dict) > 3:
            raise ValueError(
                'time_dict must contain only the keys "t0", "tend" and "dt".'
            )

        self._original_time = DMDTimeDict(dict(time_dict))
        self._dmd_time = DMDTimeDict(dict(time_dict))


def _col_major_2darray(X):
    def _col_major_2darray(X):
        """
        Converts the input snapshots into a 2D matrix by column-major ordering.

        Args:
            X: int or numpy.ndarray
                The input snapshots.

        Returns:
            snapshots: numpy.ndarray
                The 2D matrix that contains the flattened snapshots.

            snapshots_shape: tuple
                The shape of the original snapshots.
        """

    # If the data is already 2D ndarray
    if isinstance(X, np.ndarray) and X.ndim == 2:
        snapshots = X
        snapshots_shape = None
    else:
        input_shapes = [np.asarray(x).shape for x in X]

        if len(set(input_shapes)) != 1:
            raise ValueError("Snapshots have not the same dimension.")

        snapshots_shape = input_shapes[0]
        snapshots = np.transpose([np.asarray(x).flatten() for x in X])

    # check condition number of the data passed in
    cond_number = np.linalg.cond(snapshots)
    if cond_number > 10e4:
        warn(
            "Input data matrix X has condition number {}. "
            """Consider preprocessing data, passing in augmented
            data matrix, or regularization methods.""".format(
                cond_number
            )
        )

    return snapshots, snapshots_shape



from __future__ import annotations

from ._custom_observables import CustomObservables
from ._identity import Identity
from ._polynomial import Polynomial
from ._radial_basis_functions import RadialBasisFunction
from ._random_fourier_features import RandomFourierFeatures
from ._time_delay import TimeDelay

__all__ = [
    "CustomObservables",
    "Identity",
    "Polynomial",
    "RadialBasisFunction",
    "RandomFourierFeatures",
    "TimeDelay",
]



"""Wrapper classes for differentiation methods from the :doc:`derivative:index` package.

Some default values used here may differ from those used in :doc:`derivative:index`.
"""
from __future__ import annotations

from derivative import dxdt
from numpy import arange
from sklearn.base import BaseEstimator

from ..common import validate_input


class Derivative(BaseEstimator):
    """
    Wrapper class for differentiation classes from the :doc:`derivative:index` package.
    This class is meant to provide all the same functionality as the
    `dxdt <https://derivative.readthedocs.io/en/latest/api.html\
        #derivative.differentiation.dxdt>`_ method.

    This class includes a :meth:`__call__` method.

    Parameters
    ----------
    derivative_kws: dictionary, optional
        Keyword arguments to be passed to the
        `dxdt <https://derivative.readthedocs.io/en/latest/api.html\
        #derivative.differentiation.dxdt>`_
        method.

    Notes
    -----
    See the `derivative documentation <https://derivative.readthedocs.io/en/latest/>`_
    for acceptable keywords.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def set_params(self, **params):
        """
        Set the parameters of this estimator.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        else:
            self.kwargs.update(params)

        return self

    def get_params(self, deep=True):
        """Get parameters."""
        params = super().get_params(deep)

        if isinstance(self.kwargs, dict):
            params.update(self.kwargs)

        return params

    def __call__(self, x, t):
        """
        Perform numerical differentiation by calling the ``dxdt`` method.

        Paramters
        ---------
        x: np.ndarray, shape (n_samples, n_features)
            Data to be differentiated. Rows should correspond to different
            points in time and columns to different features.

        t: np.ndarray, shape (n_samples, )
            Time points for each sample (row) in ``x``.

        Returns
        -------
        x_dot: np.ndarray, shape (n_samples, n_features)
        """
        x = validate_input(x, t=t)

        if isinstance(t, (int, float)):
            if t < 0:
                raise ValueError("t must be a positive constant or an array")
            t = arange(x.shape[0]) * t

        return dxdt(x, t, axis=0, **self.kwargs)
